# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
**WARNING**

  This demo might not work as expected and some documented features might be
  missing.

-------------------------------------------------------------------------------

Creating a multi-select list box

How to use a TabularEditor to create a multi-select list box.

This demo uses two TabularEditors, side-by-side. Selections from the left table
are shown in the right table. Each table has only one column.

"""
# Issues related to the demo warning: enthought/traitsui#14,
# enthought/traitsui#15, enthought/traitsui#960

from traits.api import HasPrivateTraits, List, Str, Property
from traitsui.api import View, HGroup, UItem, TabularAdapter, TabularEditor


class MultiSelectAdapter(TabularAdapter):
    """ This adapter is used by both the left and right tables
    """

    # Titles and column names for each column of a table.
    # In this example, each table has only one column.
    columns = [('', 'myvalue')]

    # Magically named trait which gives the display text of the column named
    # 'myvalue'. This is done using a Traits Property and its getter:
    myvalue_text = Property()

    # The getter for Property 'myvalue_text' simply takes the value of the
    # corresponding item in the list being displayed in this table.
    # A more complicated example could format the item before displaying it.
    def _get_myvalue_text(self):
        return self.item


class MultiSelect(HasPrivateTraits):
    """ This is the class used to view two tables
    """
    # FIXME (TraitsUI defect #14): When multi-select is done by keyboard
    # (shift+arrow), the 'selected' trait list does not update.

    # FIXME (TraitsUI defect #15): In Windows wx, when show_titles is False,
    # left table does not draw until selection passes through all rows.
    # (Workaround here: set show_titles True and make column titles empty.)

    choices = List(Str)
    selected = List(Str)

    traits_view = View(
        HGroup(
            UItem(
                'choices',
                editor=TabularEditor(
                    show_titles=True,
                    selected='selected',
                    editable=False,
                    multi_select=True,
                    adapter=MultiSelectAdapter()
                )
            ),
            UItem(
                'selected',
                editor=TabularEditor(
                    show_titles=True,
                    editable=False,
                    adapter=MultiSelectAdapter()
                )
            )
        ),
        resizable=True,
        width=200,
        height=300
    )


# Create the demo:
demo = MultiSelect(
    choices=[
        'one', 'two', 'three', 'four', 'five', 'six', 'seven', 'eight', 'nine',
        'ten'
    ]
)

# Run the demo (if invoked from the command line):
if __name__ == '__main__':
    demo.configure_traits()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

from traitsui.qt4.check_list_editor import CustomEditor
from traitsui.testing.tester.command import MouseClick
from traitsui.testing.tester.locator import Index
from traitsui.testing.tester._ui_tester_registry._common_ui_targets import (
    BaseSourceWithLocation
)
from traitsui.testing.tester._ui_tester_registry._layout import (
    column_major_to_row_major
)
from traitsui.testing.tester._ui_tester_registry.qt4 import (
    _interaction_helpers
)


class _IndexedCustomCheckListEditor(BaseSourceWithLocation):
    """ Wrapper for CheckListEditor + locator.Index """
    source_class = CustomEditor
    locator_class = Index
    handlers = [
        (MouseClick,
            (lambda wrapper, _: _interaction_helpers.mouse_click_qlayout(
                layout=wrapper._target.source.control.layout(),
                index=convert_index(
                    layout=wrapper._target.source.control.layout(),
                    index=wrapper._target.location.index,
                ),
                delay=wrapper.delay))),
    ]


def convert_index(layout, index):
    """ Helper function to convert an index for a QGridLayout so that the
    index counts over the grid in the correct direction.
    The grid is always populated in row major order, but it is done so in
    such a way that the entries appear in column major order.
    Qlayouts are indexed in the order they are populated, so to access
    the correct element we may need to convert a column-major based index
    into a row-major one.

    Parameters
    ----------
    layout : QGridLayout
        The layout of interest
    index : int
        the index of interest
    """
    n = layout.count()
    num_cols = layout.columnCount()
    num_rows = layout.rowCount()
    return column_major_to_row_major(index, n, num_rows, num_cols)


def register(registry):
    """ Register interactions for the given registry.

    If there are any conflicts, an error will occur.

    Parameters
    ----------
    registry : TargetRegistry
        The registry being registered to.
    """
    _IndexedCustomCheckListEditor.register(registry)



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
Implementation of a BooleanEditor demo plugin for Traits UI demo program.

This demo shows each of the four styles of the BooleanEditor.

Please refer to the `BooleanEditor API docs`_ for further information.

.. _BooleanEditor API docs: https://docs.enthought.com/traitsui/api/traitsui.editors.boolean_editor.html#traitsui.editors.boolean_editor.BooleanEditor
"""

from traits.api import HasTraits, Bool
from traitsui.api import Item, Group, View


# -------------------------------------------------------------------------
#  Demo Class
# -------------------------------------------------------------------------

class BooleanEditorDemo(HasTraits):
    """ This class specifies the details of the BooleanEditor demo.
    """

    # To demonstrate any given Trait editor, an appropriate Trait is required.
    boolean_trait = Bool()

    # Items are used to define the demo display - one Item per editor style
    bool_group = Group(
        Item('boolean_trait', style='simple', label='Simple', id='simple'),
        Item('_'),
        Item('boolean_trait', style='custom', label='Custom', id='custom'),
        Item('_'),
        Item('boolean_trait', style='text', label='Text', id='text'),
        Item('_'),
        Item(
            'boolean_trait', style='readonly', label='ReadOnly', id='readonly'
        )
    )

    # Demo view
    traits_view = View(
        bool_group,
        title='BooleanEditor',
        buttons=['OK'],
        width=300
    )


# Hook for 'demo.py'
demo = BooleanEditorDemo()

# Run the demo (if invoked from the command line):
if __name__ == '__main__':
    demo.configure_traits()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" A base class for creating custom Traits UI windows.
"""


import wx

from traits.api import HasPrivateTraits, Instance, Property

from .helper import init_wx_handlers, BufferDC

# -------------------------------------------------------------------------
#  'UIWindow' class:
# -------------------------------------------------------------------------


class UIWindow(HasPrivateTraits):
    """ A base class for creating custom Traits UI windows.
    """

    #: The wx.Window associated with this custom window:
    control = Instance(wx.Window)

    #: The initial size of the window:
    size = Instance(wx.Size, (-1, -1))

    #: The current width of the window:
    width = Property()

    #: The current height of the window:
    height = Property()

    # -- Public Methods -------------------------------------------------------

    def __init__(self, parent, **traits):
        """ Creates and initializes the window.
        """
        super().__init__(**traits)
        self.control = wx.Window(
            parent, -1, size=self.size, style=wx.FULL_REPAINT_ON_RESIZE
        )
        init_wx_handlers(self.control, self)

    def refresh(self, x=None, y=None, dx=None, dy=None):
        """ Refreshes the contents of the window.
        """
        if self.control is not None:
            if x is None:
                self.control.Refresh()
            else:
                self.control.Refresh(x, y, dx, dy)

    def capture(self):
        """ Capture the mouse.
        """
        self.control.CaptureMouse()

    def release(self):
        """ Release the mouse.
        """
        self.control.ReleaseMouse()

    # -- wxPython Event Handlers ----------------------------------------------

    def _erase_background(self, event):
        """ Never, ever, do anything in this handler.
        """
        pass

    def _paint(self, event):
        """ Paints the contents of the window.
        """
        dc = BufferDC(self.control)
        self._paint_dc(dc)
        dc.copy()

    def _paint_dc(self, dc):
        """ This method should be overridden by sub-classes to do the actual
            window painting.
        """
        pass

    # -- Property Implementations ---------------------------------------------

    def _get_width(self):
        return self.control.GetClientSize()[0]

    def _set_width(self, width):
        self.control.SetSize(width, self.height)

    def _get_height(self):
        return self.control.GetClientSize()[1]

    def _set_height(self, height):
        self.control.SetSize(self.width, height)



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import unittest
from unittest.mock import Mock

from traits.api import HasTraits, Instance, Int, List, Str, Tuple

from traitsui.api import (
    Action, EvalTableFilter, Item, ObjectColumn, TableEditor, View,
)
from traitsui.tests._tools import (
    BaseTestMixin,
    create_ui,
    is_qt,
    is_wx,
    process_cascade_events,
    requires_toolkit,
    reraise_exceptions,
    ToolkitName,
)


class ListItem(HasTraits):
    """ Items to visualize in a table editor """

    value = Str()
    other_value = Int()


class ObjectListWithSelection(HasTraits):
    values = List(Instance(ListItem))
    selected = Instance(ListItem)
    selections = List(Instance(ListItem))
    selected_index = Int()
    selected_indices = List(Int)
    selected_column = Str()
    selected_columns = List(Str)
    selected_cell = Tuple(Instance(ListItem), Str)
    selected_cells = List(Tuple(Instance(ListItem), Str))
    selected_cell_index = Tuple(Int, Int)
    selected_cell_indices = List(Tuple(Int, Int))


class ObjectList(HasTraits):
    values = List(Instance(ListItem))


simple_view = View(
    Item(
        "values",
        show_label=False,
        editor=TableEditor(
            columns=[
                ObjectColumn(name="value"),
                ObjectColumn(name="other_value"),
            ]
        ),
    ),
    buttons=["OK"],
)

filtered_view = View(
    Item(
        "values",
        show_label=False,
        editor=TableEditor(
            columns=[
                ObjectColumn(name="value"),
                ObjectColumn(name="other_value"),
            ],
            filter=EvalTableFilter(expression="other_value >= 2"),
        ),
    ),
    buttons=["OK"],
)

select_row_view = View(
    Item(
        "values",
        show_label=False,
        editor=TableEditor(
            columns=[
                ObjectColumn(name="value"),
                ObjectColumn(name="other_value"),
            ],
            selection_mode="row",
            selected="selected",
        ),
    ),
    buttons=["OK"],
)

select_rows_view = View(
    Item(
        "values",
        show_label=False,
        editor=TableEditor(
            columns=[ObjectColumn(name="value")],
            selection_mode="rows",
            selected="selections",
        ),
    ),
    buttons=["OK"],
)

select_row_index_view = View(
    Item(
        "values",
        show_label=False,
        editor=TableEditor(
            columns=[
                ObjectColumn(name="value"),
                ObjectColumn(name="other_value"),
            ],
            selection_mode="row",
            selected_indices="selected_index",
        ),
    ),
    buttons=["OK"],
)

select_row_indices_view = View(
    Item(
        "values",
        show_label=False,
        editor=TableEditor(
            columns=[
                ObjectColumn(name="value"),
                ObjectColumn(name="other_value"),
            ],
            selection_mode="rows",
            selected_indices="selected_indices",
        ),
    ),
    buttons=["OK"],
)

select_column_view = View(
    Item(
        "values",
        show_label=False,
        editor=TableEditor(
            columns=[
                ObjectColumn(name="value"),
                ObjectColumn(name="other_value"),
            ],
            selection_mode="column",
            selected="selected_column",
        ),
    ),
    buttons=["OK"],
)

select_columns_view = View(
    Item(
        "values",
        show_label=False,
        editor=TableEditor(
            columns=[
                ObjectColumn(name="value"),
                ObjectColumn(name="other_value"),
            ],
            selection_mode="columns",
            selected="selected_columns",
        ),
    ),
    buttons=["OK"],
)

select_column_index_view = View(
    Item(
        "values",
        show_label=False,
        editor=TableEditor(
            columns=[
                ObjectColumn(name="value"),
                ObjectColumn(name="other_value"),
            ],
            selection_mode="column",
            selected_indices="selected_index",
        ),
    ),
    buttons=["OK"],
)

select_column_indices_view = View(
    Item(
        "values",
        show_label=False,
        editor=TableEditor(
            columns=[
                ObjectColumn(name="value"),
                ObjectColumn(name="other_value"),
            ],
            selection_mode="columns",
            selected_indices="selected_indices",
        ),
    ),
    buttons=["OK"],
)

select_cell_view = View(
    Item(
        "values",
        show_label=False,
        editor=TableEditor(
            columns=[
                ObjectColumn(name="value"),
                ObjectColumn(name="other_value"),
            ],
            selection_mode="cell",
            selected="selected_cell",
        ),
    ),
    buttons=["OK"],
)

select_cells_view = View(
    Item(
        "values",
        show_label=False,
        editor=TableEditor(
            columns=[
                ObjectColumn(name="value"),
                ObjectColumn(name="other_value"),
            ],
            selection_mode="cells",
            selected="selected_cells",
        ),
    ),
    buttons=["OK"],
)

select_cell_index_view = View(
    Item(
        "values",
        show_label=False,
        editor=TableEditor(
            columns=[
                ObjectColumn(name="value"),
                ObjectColumn(name="other_value"),
            ],
            selection_mode="cell",
            selected_indices="selected_cell_index",
        ),
    ),
    buttons=["OK"],
)

select_cell_indices_view = View(
    Item(
        "values",
        show_label=False,
        editor=TableEditor(
            columns=[
                ObjectColumn(name="value"),
                ObjectColumn(name="other_value"),
            ],
            selection_mode="cells",
            selected_indices="selected_cell_indices",
        ),
    ),
    buttons=["OK"],
)


class TestTableEditor(BaseTestMixin, unittest.TestCase):

    def setUp(self):
        BaseTestMixin.setUp(self)

    def tearDown(self):
        BaseTestMixin.tearDown(self)

    @requires_toolkit([ToolkitName.qt, ToolkitName.wx])
    def test_table_editor(self):
        object_list = ObjectListWithSelection(
            values=[ListItem(value=str(i ** 2)) for i in range(10)]
        )

        with reraise_exceptions(), \
                create_ui(object_list, dict(view=simple_view)):
            process_cascade_events()

    @requires_toolkit([ToolkitName.qt, ToolkitName.wx])
    def test_filtered_table_editor(self):
        object_list = ObjectListWithSelection(
            values=[ListItem(value=str(i ** 2)) for i in range(10)]
        )

        with reraise_exceptions(), \
                create_ui(object_list, dict(view=filtered_view)) as ui:
            process_cascade_events()

            filter = ui.get_editors("values")[0].filter

            process_cascade_events()

        self.assertIsNotNone(filter)

    @requires_toolkit([ToolkitName.qt, ToolkitName.wx])
    def test_table_editor_select_row(self):
        object_list = ObjectListWithSelection(
            values=[ListItem(value=str(i ** 2)) for i in range(10)]
        )
        object_list.selected = object_list.values[5]

        with reraise_exceptions(), \
                create_ui(object_list, dict(view=select_row_view)) as ui:
            editor = ui.get_editors("values")[0]
            process_cascade_events()
            if is_qt():
                selected = editor.selected
            elif is_wx():
                selected = editor.selected_row

            process_cascade_events()

        self.assertIs(selected, object_list.values[5])

    @requires_toolkit([ToolkitName.qt, ToolkitName.wx])
    def test_table_editor_select_rows(self):
        object_list = ObjectListWithSelection(
            values=[ListItem(value=str(i ** 2)) for i in range(10)]
        )
        object_list.selections = object_list.values[5:7]

        with reraise_exceptions(), \
                create_ui(object_list, dict(view=select_rows_view)) as ui:
            editor = ui.get_editors("values")[0]
            process_cascade_events()
            if is_qt():
                selected = editor.selected
            elif is_wx():
                selected = editor.selected_rows

            process_cascade_events()

        self.assertEqual(selected, object_list.values[5:7])

    @requires_toolkit([ToolkitName.qt, ToolkitName.wx])
    def test_table_editor_select_row_index(self):
        object_list = ObjectListWithSelection(
            values=[ListItem(value=str(i ** 2)) for i in range(10)]
        )
        object_list.selected_index = 5

        with reraise_exceptions(), \
                create_ui(object_list, dict(view=select_row_index_view)) as ui:
            editor = ui.get_editors("values")[0]
            process_cascade_events()
            if is_qt():
                selected = editor.selected_indices
            elif is_wx():
                selected = editor.selected_row_index

            process_cascade_events()

        self.assertEqual(selected, 5)

    @requires_toolkit([ToolkitName.qt, ToolkitName.wx])
    def test_table_editor_select_row_indices(self):
        object_list = ObjectListWithSelection(
            values=[ListItem(value=str(i ** 2)) for i in range(10)]
        )
        object_list.selected_indices = [5, 7, 8]

        view = select_row_indices_view
        with reraise_exceptions(), \
                create_ui(object_list, dict(view=view)) as ui:
            editor = ui.get_editors("values")[0]
            process_cascade_events()
            if is_qt():
                selected = editor.selected_indices
            elif is_wx():
                selected = editor.selected_row_indices

            process_cascade_events()

        self.assertEqual(selected, [5, 7, 8])

    @requires_toolkit([ToolkitName.qt, ToolkitName.wx])
    def test_table_editor_select_column(self):
        object_list = ObjectListWithSelection(
            values=[ListItem(value=str(i ** 2)) for i in range(10)]
        )
        object_list.selected_column = "value"

        with reraise_exceptions(), \
                create_ui(object_list, dict(view=select_column_view)) as ui:
            editor = ui.get_editors("values")[0]
            process_cascade_events()
            if is_qt():
                selected = editor.selected
            elif is_wx():
                selected = editor.selected_column

            process_cascade_events()

        self.assertEqual(selected, "value")

    @requires_toolkit([ToolkitName.qt, ToolkitName.wx])
    def test_table_editor_select_columns(self):
        object_list = ObjectListWithSelection(
            values=[ListItem(value=str(i ** 2)) for i in range(10)]
        )
        object_list.selected_columns = ["value", "other_value"]

        with reraise_exceptions(), \
                create_ui(object_list, dict(view=select_columns_view)) as ui:
            editor = ui.get_editors("values")[0]
            process_cascade_events()
            if is_qt():
                selected = editor.selected
            elif is_wx():
                selected = editor.selected_columns

            process_cascade_events()

        self.assertEqual(selected, ["value", "other_value"])

    @requires_toolkit([ToolkitName.qt, ToolkitName.wx])
    def test_table_editor_select_column_index(self):
        object_list = ObjectListWithSelection(
            values=[ListItem(value=str(i ** 2)) for i in range(10)]
        )
        object_list.selected_index = 1

        view = select_column_index_view
        with reraise_exceptions(), \
                create_ui(object_list, dict(view=view)) as ui:
            editor = ui.get_editors("values")[0]
            process_cascade_events()
            if is_qt():
                selected = editor.selected_indices
            elif is_wx():
                selected = editor.selected_column_index

            process_cascade_events()

        self.assertEqual(selected, 1)

    @requires_toolkit([ToolkitName.qt, ToolkitName.wx])
    def test_table_editor_select_column_indices(self):
        object_list = ObjectListWithSelection(
            values=[ListItem(value=str(i ** 2)) for i in range(10)]
        )
        object_list.selected_indices = [0, 1]

        view = select_column_indices_view
        with reraise_exceptions(), \
                create_ui(object_list, dict(view=view)) as ui:
            editor = ui.get_editors("values")[0]
            process_cascade_events()
            if is_qt():
                selected = editor.selected_indices
            elif is_wx():
                selected = editor.selected_column_indices

            process_cascade_events()

        self.assertEqual(selected, [0, 1])

    @requires_toolkit([ToolkitName.qt, ToolkitName.wx])
    def test_table_editor_select_cell(self):
        object_list = ObjectListWithSelection(
            values=[ListItem(value=str(i ** 2)) for i in range(10)]
        )
        object_list.selected_cell = (object_list.values[5], "value")

        with reraise_exceptions(), \
                create_ui(object_list, dict(view=select_cell_view)) as ui:
            editor = ui.get_editors("values")[0]
            process_cascade_events()
            if is_qt():
                selected = editor.selected
            elif is_wx():
                selected = editor.selected_cell

            process_cascade_events()

        self.assertEqual(selected, (object_list.values[5], "value"))

    @requires_toolkit([ToolkitName.qt, ToolkitName.wx])
    def test_table_editor_select_cells(self):
        object_list = ObjectListWithSelection(
            values=[ListItem(value=str(i ** 2)) for i in range(10)]
        )
        object_list.selected_cells = [
            (object_list.values[5], "value"),
            (object_list.values[6], "other value"),
            (object_list.values[8], "value"),
        ]

        with reraise_exceptions(), \
                create_ui(object_list, dict(view=select_cells_view)) as ui:
            editor = ui.get_editors("values")[0]
            process_cascade_events()
            if is_qt():
                selected = editor.selected
            elif is_wx():
                selected = editor.selected_cells

            process_cascade_events()

        self.assertEqual(selected, [
            (object_list.values[5], "value"),
            (object_list.values[6], "other value"),
            (object_list.values[8], "value"),
        ])

    @requires_toolkit([ToolkitName.qt, ToolkitName.wx])
    def test_table_editor_select_cell_index(self):
        object_list = ObjectListWithSelection(
            values=[ListItem(value=str(i ** 2)) for i in range(10)]
        )
        object_list.selected_cell_index = (5, 1)

        view = select_cell_index_view
        with reraise_exceptions(), \
                create_ui(object_list, dict(view=view)) as ui:
            editor = ui.get_editors("values")[0]
            process_cascade_events()
            if is_qt():
                selected = editor.selected_indices
            elif is_wx():
                selected = editor.selected_cell_index

            process_cascade_events()

        self.assertEqual(selected, (5, 1))

    @requires_toolkit([ToolkitName.qt, ToolkitName.wx])
    def test_table_editor_select_cell_indices(self):
        object_list = ObjectListWithSelection(
            values=[ListItem(value=str(i ** 2)) for i in range(10)]
        )
        object_list.selected_cell_indices = [(5, 0), (6, 1), (8, 0)]

        view = select_cell_indices_view
        with reraise_exceptions(), \
                create_ui(object_list, dict(view=view)) as ui:
            editor = ui.get_editors("values")[0]
            process_cascade_events()
            if is_qt():
                selected = editor.selected_indices
            elif is_wx():
                selected = editor.selected_cell_indices

            process_cascade_events()

        self.assertEqual(selected, [(5, 0), (6, 1), (8, 0)])

    @requires_toolkit([ToolkitName.qt])
    def test_progress_column(self):
        from traitsui.extras.progress_column import ProgressColumn

        progress_view = View(
            Item(
                "values",
                show_label=False,
                editor=TableEditor(
                    columns=[
                        ObjectColumn(name="value"),
                        ProgressColumn(name="other_value"),
                    ]
                ),
            ),
            buttons=["OK"],
        )
        object_list = ObjectList(
            values=[ListItem(value=str(i ** 2)) for i in range(10)]
        )

        with reraise_exceptions(), \
                create_ui(object_list, dict(view=progress_view)):
            process_cascade_events()

    @requires_toolkit([ToolkitName.qt])
    def test_on_perform_action(self):
        # A test for issue #741, where actions with an on_perform function set
        # would get called twice
        object_list = ObjectList(
            values=[ListItem(value=str(i ** 2)) for i in range(10)]
        )
        mock_function = Mock()
        action = Action(on_perform=mock_function)

        with reraise_exceptions(), \
                create_ui(object_list, dict(view=simple_view)) as ui:
            editor = ui.get_editors("values")[0]
            process_cascade_events()
            editor.set_menu_context(None, None, None)
            editor.perform(action)
        mock_function.assert_called_once()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Defines the array editor factory for all traits toolkit backends.
"""

import numpy

from traits.api import Bool, HasTraits, Int, Float, Instance, TraitError

from traitsui.editor import Editor
from traitsui.editor_factory import EditorFactory
from traitsui.group import Group
from traitsui.item import Item
from traitsui.view import View


class ArrayEditor(EditorFactory):
    """ Editor factory for array editors.
    """

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: Width of the individual fields
    width = Int(-80)

    #: Is user input set on every keystroke?
    auto_set = Bool(True)

    #: Is user input set when the Enter key is pressed?
    enter_set = Bool(False)


class ArrayStructure(HasTraits):

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: Editor that this structure is linked to
    editor = Instance(Editor)

    #: The constructed View for the array
    view = Instance(View)

    def __init__(self, editor):
        """ Initializes the object.
        """
        # Save the reference to the editor:
        self.editor = editor

        # Set up the field width for each item:
        width = editor.factory.width

        # Set up the correct style for each filed:
        style = "simple"
        if editor.readonly:
            style = "readonly"

        # Get the array we are mirroring:
        object = editor.value

        # Determine the correct trait type to use for each element:
        trait = Float()

        if object.dtype.type == "i":
            trait = Int()

        if len(object.shape) == 1:
            self.view = self._one_dim_view(object, style, width, trait)
        elif len(object.shape) == 2:
            self.view = self._two_dim_view(object, style, width, trait)
        else:
            raise TraitError("Only 1D or 2D arrays supported")

    # -------------------------------------------------------------------------
    #  1D view:
    # -------------------------------------------------------------------------

    def _one_dim_view(self, object, style, width, trait):
        content = []
        shape = object.shape
        items = []
        format_func = self.editor.factory.format_func
        format_str = self.editor.factory.format_str
        for i in range(shape[0]):
            name = "f%d" % i
            self.add_trait(
                name,
                trait(
                    object[i],
                    event="field",
                    auto_set=self.editor.factory.auto_set,
                    enter_set=self.editor.factory.enter_set,
                ),
            )
            items.append(
                Item(
                    name=name,
                    style=style,
                    width=width,
                    format_func=format_func,
                    format_str=format_str,
                    padding=-3,
                )
            )

        group = Group(orientation="horizontal", show_labels=False, *items)
        content.append(group)

        return View(Group(show_labels=False, *content))

    # -------------------------------------------------------------------------
    #  2D view:
    # -------------------------------------------------------------------------

    def _two_dim_view(self, object, style, width, trait):
        content = []
        shape = object.shape
        format_func = self.editor.factory.format_func
        format_str = self.editor.factory.format_str
        for i in range(shape[0]):
            items = []
            for j in range(shape[1]):
                name = "f%d_%d" % (i, j)
                self.add_trait(
                    name,
                    trait(
                        object[i, j],
                        event="field",
                        auto_set=self.editor.factory.auto_set,
                        enter_set=self.editor.factory.enter_set,
                    ),
                )
                items.append(
                    Item(
                        name=name,
                        style=style,
                        width=width,
                        format_func=format_func,
                        format_str=format_str,
                        padding=-3,
                    )
                )

            group = Group(orientation="horizontal", show_labels=False, *items)
            content.append(group)

        return View(Group(show_labels=False, *content))

    def _field_changed(self):
        """ Updates the underlying array when any field changes value.
        """

        if not self.editor._busy:
            # Get the array we are mirroring:
            object = self.editor.value
            shape = object.shape
            value = numpy.zeros(shape, object.dtype)

            # 1D
            if len(shape) == 1:
                for i in range(shape[0]):
                    value[i] = getattr(self, "f%d" % i)
            # 2D
            elif len(shape) == 2:
                for i in range(shape[0]):
                    for j in range(shape[1]):
                        value[i, j] = getattr(self, "f%d_%d" % (i, j))

            self.editor.update_array(value)


# -------------------------------------------------------------------------
#  Toolkit-independent 'SimpleEditor' class:
# -------------------------------------------------------------------------


class SimpleEditor(Editor):
    """ Simple style of editor for arrays.
    """

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: Is the editor read-only?
    readonly = Bool(False)

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        self._as = _as = ArrayStructure(self)
        ui = _as.view.ui(_as, parent, kind="subpanel")
        ui.parent = self.ui
        self.control = ui.control

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """

        if not self._busy:
            self._busy = True
            object = self.value
            shape = object.shape
            _as = self._as

            # 1D
            if len(shape) == 1:
                for i in range(shape[0]):
                    setattr(_as, "f%d" % i, object[i])
            # 2D
            elif len(shape) == 2:
                for i in range(shape[0]):
                    for j in range(shape[1]):
                        setattr(_as, "f%d_%d" % (i, j), object[i, j])

            self._busy = False

    def update_array(self, value):
        """ Updates the array value associated with the editor.
        """
        self._busy = True
        self.value = value
        self._busy = False


# This alias is deprecated and will be removed in TraitsUI 8.
ToolkitEditorFactory = ArrayEditor



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

from traitsui.wx.boolean_editor import ReadonlyEditor, SimpleEditor
from traitsui.testing.tester.command import MouseClick
from traitsui.testing.tester.query import DisplayedText, IsChecked
from traitsui.testing.tester._ui_tester_registry.wx import _interaction_helpers


def register(registry):
    """ Register solvers/handlers specific to wx Boolean Editors
    for the given registry.

    If there are any conflicts, an error will occur.

    Parameters
    ----------
    registry : TargetRegistry
    """
    registry.register_interaction(
        target_class=SimpleEditor,
        interaction_class=MouseClick,
        handler=(lambda wrapper, _: _interaction_helpers.mouse_click_checkbox(
                 control=wrapper._target.control, delay=wrapper.delay))
    )

    registry.register_interaction(
        target_class=SimpleEditor,
        interaction_class=IsChecked,
        handler=lambda wrapper, _: wrapper._target.control.GetValue()
    )

    registry.register_interaction(
        target_class=ReadonlyEditor,
        interaction_class=DisplayedText,
        handler=lambda wrapper, _:
            _interaction_helpers.readonly_textbox_displayed_text(
                control=wrapper._target.control
            )
    )



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Defines the table column descriptor used for editing the object represented
    by the row
"""



from ..table_column import ObjectColumn


class EditColumn(ObjectColumn):
    def __init__(self, **traits):
        """ Initializes the object.
        """
        super().__init__(**traits)

        from traitsui.toolkit import toolkit_object
        EditRenderer = toolkit_object(
            'extra.edit_renderer:EditRenderer'
        )
        self.renderer = EditRenderer()

        self.label = ""

    def get_cell_color(self, object):
        """ Returns the cell background color for the column for a specified
            object.
        """

        # Override the parent class to ALWAYS provide the standard color:
        return self.cell_color_

    def is_editable(self, object):
        """ Returns whether the column is editable for a specified object.
        """
        return False



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from traits.api \
    import HasTraits, List

from traitsui.api \
    import View, Item, TableEditor

from traitsui.color_column \
    import ColorColumn

from enable.api \
    import ColorTrait


class Thingy(HasTraits):
    color = ColorTrait('black')

#-------------------------------------------------------------------------
#  Sample data:
#-------------------------------------------------------------------------

colors = [
    Thingy(color='red'),
    Thingy(color='orange'),
    Thingy(color='yellow'),
    Thingy(color='green'),
    Thingy(color='blue'),
    Thingy(color='indigo'),
    Thingy(color='violet'),
    Thingy(color='black'),
    Thingy(color='white'),
]


class TableTest(HasTraits):

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    colors = List(Thingy)

    table_editor = TableEditor(
        columns=[ColorColumn(name='color'),
                 ],

        editable=True,
        deletable=True,
        sortable=True,        #
        sort_model=True,
        show_lines=True,        #
        orientation='vertical',
        show_column_labels=True,        #
        row_factory=Thingy
    )

    traits_view = View(
        [Item('colors',
              id='colors',
              editor=table_editor),
         '|[]<>'],
        title='Table Editor Test',
        id='traitsui.tests.table_editor_color_test',
        dock='horizontal',
        width=.4,
        height=.3,
        resizable=True,
        kind='live')

#-------------------------------------------------------------------------
#  Run the tests:
#-------------------------------------------------------------------------

if __name__ == '__main__':
    tt = TableTest(colors=colors)
    tt.configure_traits()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import io
import textwrap
import unittest
from unittest import mock

from pyface.api import GUI

from traits.api import HasTraits, Int
from traits.testing.api import UnittestTools
from traitsui.tests._tools import (
    process_cascade_events,
    requires_toolkit,
    ToolkitName,
)
from traitsui.testing.tester._abstract_target_registry import (
    AbstractTargetRegistry,
)
from traitsui.testing.tester.exceptions import (
    InteractionNotSupported,
    LocationNotSupported,
)
from traitsui.testing.tester.target_registry import (
    TargetRegistry,
)
from traitsui.testing.tester.ui_wrapper import (
    UIWrapper,
)


def example_ui_wrapper(**kwargs):
    """ Return an instance of UIWrapper for testing purposes.

    Parameters
    ----------
    **kwargs
        Values to use instead of the default values.

    Returns
    -------
    wrapper: UIWrapper
    """
    values = dict(
        target=None,
        registries=[],
    )
    values.update(kwargs)
    return UIWrapper(**values)


class StubRegistry(AbstractTargetRegistry):
    """ A stub implementation of the AbstractTargetRegistry for testing
    """

    def __init__(
            self,
            handler=None,
            solver=None,
            supported_interaction_classes=(),
            supported_locator_classes=(),
            interaction_doc="",
            location_doc="",
            ):
        self.handler = handler
        self.solver = solver
        self.supported_interaction_classes = supported_interaction_classes
        self.supported_locator_classes = supported_locator_classes
        self.interaction_doc = interaction_doc
        self.location_doc = location_doc

    def _get_handler(self, target, interaction):
        if interaction.__class__ not in self.supported_interaction_classes:
            raise InteractionNotSupported(
                target_class=target.__class__,
                interaction_class=interaction.__class__,
                supported=list(self.supported_interaction_classes),
            )
        return self.handler

    def _get_interactions(self, target):
        return set(self.supported_interaction_classes)

    def _get_interaction_doc(self, target, interaction_class):
        return self.interaction_doc

    def _get_solver(self, target, location):
        if location.__class__ not in self.supported_locator_classes:
            raise LocationNotSupported(
                target_class=target.__class__,
                locator_class=location.__class__,
                supported=list(self.supported_locator_classes),
            )
        return self.solver

    def _get_locations(self, target):
        return set(self.supported_locator_classes)

    def _get_location_doc(self, target, locator_class):
        return self.location_doc


# Use of perform/inspect requires the GUI event loop
@requires_toolkit([ToolkitName.qt, ToolkitName.wx])
class TestUIWrapperInteractionRegistries(unittest.TestCase):
    """ Test the logic regarding the order of (interaction) registries.
    """

    def test_registry_priority(self):
        # If two registries have a handler for the same target and interaction
        # types, the first register is used.
        registry1 = StubRegistry(
            handler=lambda w, l: 1,
            supported_interaction_classes=[str],
        )
        registry2 = StubRegistry(
            handler=lambda w, l: 2,
            supported_interaction_classes=[str],
        )

        wrapper = example_ui_wrapper(
            registries=[registry2, registry1],
        )
        value = wrapper.inspect("some string")

        self.assertEqual(value, 2)

        # reverse order
        wrapper = example_ui_wrapper(
            registries=[registry1, registry2]
        )
        value = wrapper.inspect("some other string")

        self.assertEqual(value, 1)

    def test_registry_selection(self):
        # If the first registry says it can't handle the interaction, the next
        # registry is tried.

        registry1 = StubRegistry()
        registry2_handler = mock.Mock()
        registry2 = StubRegistry(
            handler=registry2_handler,
            supported_interaction_classes=[int],
        )

        wrapper = example_ui_wrapper(
            registries=[registry1, registry2],
        )
        wrapper.perform(123)

        self.assertEqual(registry2_handler.call_count, 1)

    def test_registry_all_declined(self):
        # If none of the registries can support an interaction, the
        # exception raised provide information on what actions are
        # supported.

        wrapper = example_ui_wrapper(
            registries=[
                StubRegistry(supported_interaction_classes=[int]),
                StubRegistry(supported_interaction_classes=[float]),
            ],
        )
        with self.assertRaises(InteractionNotSupported) as exception_context:
            wrapper.perform(None)

        self.assertCountEqual(
            exception_context.exception.supported,
            [int, float],
        )


# Use of locate requires the GUI event loop
@requires_toolkit([ToolkitName.qt, ToolkitName.wx])
class TestUIWrapperLocationRegistry(unittest.TestCase):
    """ Test the use of registries with locate. """

    def test_location_registry_priority(self):
        registry1 = StubRegistry(
            solver=lambda w, l: 1,
            supported_locator_classes=[str],
        )
        registry2 = StubRegistry(
            solver=lambda w, l: 2,
            supported_locator_classes=[str],
        )

        wrapper = example_ui_wrapper(
            registries=[registry2, registry1],
        )
        new_wrapper = wrapper.locate("some string")

        self.assertEqual(new_wrapper._target, 2)

        # swap the order
        wrapper = example_ui_wrapper(
            registries=[registry1, registry2],
        )
        new_wrapper = wrapper.locate("some other string")

        self.assertEqual(new_wrapper._target, 1)

    def test_location_registry_selection(self):
        # If the first registry says it can't handle the interaction, the next
        # registry is tried.

        def solver2(wrapper, location):
            return 2

        registry1 = StubRegistry()
        registry2 = StubRegistry(
            solver=solver2,
            supported_locator_classes=[str],
        )

        wrapper = example_ui_wrapper(
            registries=[registry1, registry2],
        )
        new_wrapper = wrapper.locate("some string")

        self.assertEqual(new_wrapper._target, 2)
        self.assertEqual(
            new_wrapper._registries,
            wrapper._registries,
        )

    def test_registry_all_declined(self):
        # If none of the registries can support a location, the
        # exception raised provide information on what actions are
        # supported.

        wrapper = example_ui_wrapper(
            registries=[
                StubRegistry(supported_locator_classes=[int]),
                StubRegistry(supported_locator_classes=[float]),
            ],
        )
        with self.assertRaises(LocationNotSupported) as exception_context:
            wrapper.locate(None)

        self.assertCountEqual(
            exception_context.exception.supported,
            [int, float],
        )


class TestUIWrapperHelp(unittest.TestCase):
    """ Test calling UIWrapper.help """

    def test_help_message(self):

        class Action:
            """ Say hello.
            Say bye.
            """
            pass

        class Locator:
            """ Return anything you like.
            Good day!
            """
            pass

        registry1 = TargetRegistry()
        registry1.register_interaction(
            target_class=str,
            interaction_class=Action,
            handler=mock.Mock(),
        )
        registry2 = TargetRegistry()
        registry2.register_location(
            target_class=str,
            locator_class=Locator,
            solver=mock.Mock(),
        )

        wrapper = example_ui_wrapper(
            target="dummy", registries=[registry1, registry2]
        )

        # when
        stream = io.StringIO()
        with mock.patch("sys.stdout", stream):
            wrapper.help()

        # then
        self.assertEqual(
            stream.getvalue(),
            textwrap.dedent(f"""\
                Interactions
                ------------
                {Action!r}
                    Say hello.
                    Say bye.

                Locations
                ---------
                {Locator!r}
                    Return anything you like.
                    Good day!

            """)
        )

    def test_help_message_priority_interactions(self):
        # The first registry in the list has the highest priority
        # The last registry in the list has the least priority

        high_priority_registry = StubRegistry(
            supported_locator_classes=[str],
            supported_interaction_classes=[float],
            interaction_doc="Interaction: I get a higher priority.",
            location_doc="Location: I get a higher priority.",
        )

        low_priority_registry = StubRegistry(
            supported_locator_classes=[str],
            supported_interaction_classes=[float],
            interaction_doc="Interaction: I get a lower priority.",
            location_doc="Location: I get a lower priority.",
        )

        # Put higher priority registry first.
        wrapper = example_ui_wrapper(
            target="dummy",
            registries=[high_priority_registry, low_priority_registry],
        )

        # when
        stream = io.StringIO()
        with mock.patch("sys.stdout", stream):
            wrapper.help()

        # then
        self.assertEqual(
            stream.getvalue(),
            textwrap.dedent(f"""\
                Interactions
                ------------
                {float!r}
                    Interaction: I get a higher priority.

                Locations
                ---------
                {str!r}
                    Location: I get a higher priority.

            """)
        )

    def test_help_message_nothing_is_supported(self):
        registry = TargetRegistry()
        wrapper = example_ui_wrapper(registries=[registry])

        # when
        stream = io.StringIO()
        with mock.patch("sys.stdout", stream):
            wrapper.help()

        # then
        self.assertEqual(
            stream.getvalue(),
            textwrap.dedent("""\
                Interactions
                ------------
                No interactions are supported.

                Locations
                ---------
                No locations are supported.

            """)
        )


class NumberHasTraits(HasTraits):
    number = Int()


@requires_toolkit([ToolkitName.qt, ToolkitName.wx])
class TestUIWrapperEventProcessed(unittest.TestCase, UnittestTools):
    """ Test GUI events are processed and exceptions from the GUI event
    loop are handled.
    """

    def test_event_processed(self):
        # Test GUI events are processed such that the trait is changed.
        gui = GUI()
        model = NumberHasTraits()

        def handler(wrapper, action):
            gui.set_trait_later(model, "number", 2)

        wrapper = example_ui_wrapper(
            registries=[
                StubRegistry(
                    handler=handler,
                    supported_interaction_classes=[float],
                ),
            ],
        )

        with self.assertTraitChanges(model, "number"):
            wrapper.perform(2.123)

    def test_event_processed_prior_to_resolving_location(self):
        # Test GUI events are processed prior to resolving location
        gui = GUI()
        model = NumberHasTraits()
        gui.set_trait_later(model, "number", 2)

        def solver(wrapper, location):
            return model.number

        wrapper = example_ui_wrapper(
            registries=[
                StubRegistry(
                    solver=solver,
                    supported_locator_classes=[float],
                )
            ],
        )

        new_wrapper = wrapper.locate(4.567)
        self.assertEqual(new_wrapper._target, 2)

    def test_event_processed_with_exception_captured(self):
        # Test exceptions in the GUI event loop are captured and then cause
        # the test to error.
        gui = GUI()

        def raise_error():
            raise ZeroDivisionError()

        def handler(wrapper, action):
            gui.invoke_later(raise_error)

        wrapper = example_ui_wrapper(
            registries=[
                StubRegistry(
                    handler=handler,
                    supported_interaction_classes=[float],
                ),
            ],
        )
        with self.assertRaises(RuntimeError), self.assertLogs("traitsui"):
            wrapper.perform(1.234)

    def test_exception_not_in_gui(self):
        # Exceptions from code executed outside of the event loop are
        # propagated as is.

        def handler(wrapper, action):
            raise ZeroDivisionError()

        wrapper = example_ui_wrapper(
            registries=[
                StubRegistry(
                    handler=handler,
                    supported_interaction_classes=[float],
                ),
            ],
        )

        with self.assertRaises(ZeroDivisionError):
            wrapper.perform(9.99)

    def test_perform_event_processed_optional(self):
        # Allow event processing to be switched off.
        gui = GUI()
        side_effect = mock.Mock()

        def handler(wrapper, action):
            gui.invoke_later(side_effect)

        wrapper = example_ui_wrapper(
            registries=[
                StubRegistry(
                    handler=handler,
                    supported_interaction_classes=[float],
                ),
            ],
            auto_process_events=False,
        )

        # With auto_process_events set to False, events are not automatically
        # processed.
        wrapper.perform(12.345)
        self.addCleanup(process_cascade_events)

        self.assertEqual(side_effect.call_count, 0)

    def test_locate_event_processed_optional(self):
        # Allow event processing to be switched off.
        gui = GUI()
        side_effect = mock.Mock()
        gui.invoke_later(side_effect)
        self.addCleanup(process_cascade_events)

        def solver(wrapper, location):
            return 1

        wrapper = example_ui_wrapper(
            registries=[
                StubRegistry(
                    solver=solver,
                    supported_locator_classes=[float],
                ),
            ],
            auto_process_events=False,
        )

        # With auto_process_events set to False, events are not automatically
        # processed.
        new_wrapper = wrapper.locate(1.234)

        self.assertEqual(side_effect.call_count, 0)
        self.assertFalse(new_wrapper._auto_process_events)



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import contextlib
import unittest

from traits.api import HasTraits, List, Str
from traitsui.api import CheckListEditor, UItem, View
from traitsui.tests._tools import (
    BaseTestMixin,
    create_ui,
    get_all_button_status,
    is_qt,
    is_wx,
    process_cascade_events,
    requires_toolkit,
    reraise_exceptions,
    ToolkitName,
)
from traitsui.testing.api import (
    Index,
    MouseClick,
    UITester
)


class ListModel(HasTraits):

    value = List()


def get_view(style):
    return View(
        UItem(
            "value",
            editor=CheckListEditor(
                values=["one", "two", "three", "four"],
            ),
            style=style,
        ),
        resizable=True
    )


def get_view_custom_cols(cols):
    return View(
        UItem(
            "value",
            editor=CheckListEditor(
                values=["one", "two", "three", "four", "five", "six", "seven"],
                cols=cols,
            ),
            style="custom",
        ),
        resizable=True
    )


def get_mapped_view(style):
    return View(
        UItem(
            "value",
            editor=CheckListEditor(
                values=[(1, "one"), (2, "two"), (3, "three"), (4, "four")],
            ),
            style=style,
        ),
        resizable=True
    )


def get_combobox_text(combobox):
    """ Return the text given a combobox control. """
    if is_wx():
        return combobox.GetString(combobox.GetSelection())

    elif is_qt():
        return combobox.currentText()

    else:
        raise unittest.SkipTest("Test not implemented for this toolkit")


def set_combobox_index(editor, idx):
    """ Set the choice index of a combobox control given editor and index
    number. """
    if is_wx():
        import wx

        choice = editor.control
        choice.SetSelection(idx)
        event = wx.CommandEvent(wx.EVT_CHOICE.typeId, choice.GetId())
        event.SetString(choice.GetString(idx))
        wx.PostEvent(choice, event)

    elif is_qt():
        # Cannot initiate update programatically because of `activated`
        # event. At least check that it updates as expected when done
        # manually
        editor.update_object(idx)

    else:
        raise unittest.SkipTest("Test not implemented for this toolkit")


def set_text_in_line_edit(line_edit, text):
    """ Set text in text widget and complete editing. """
    if is_wx():
        import wx

        line_edit.SetValue(text)
        event = wx.CommandEvent(wx.EVT_TEXT_ENTER.typeId, line_edit.GetId())
        wx.PostEvent(line_edit, event)

    elif is_qt():
        line_edit.setText(text)
        line_edit.editingFinished.emit()

    else:
        raise unittest.SkipTest("Test not implemented for this toolkit")


@requires_toolkit([ToolkitName.qt, ToolkitName.wx])
class TestCheckListEditorMapping(BaseTestMixin, unittest.TestCase):

    def setUp(self):
        BaseTestMixin.setUp(self)

    def tearDown(self):
        BaseTestMixin.tearDown(self)

    @contextlib.contextmanager
    def setup_ui(self, model, view):
        with create_ui(model, dict(view=view)) as ui:
            yield ui.get_editors("value")[0]

    def check_checklist_mappings_value_change(self, style):
        check_list_editor_factory = CheckListEditor(
            values=["one", "two"],
            format_func=lambda v: v.upper(),
        )
        formatted_view = View(
            UItem(
                "value",
                editor=check_list_editor_factory,
                style=style,
            )
        )
        model = ListModel()

        with reraise_exceptions(), \
                self.setup_ui(model, formatted_view) as editor:

            self.assertEqual(editor.names, ["ONE", "TWO"])

            check_list_editor_factory.values = ["two", "one"]

            self.assertEqual(editor.names, ["TWO", "ONE"])

    def check_checklist_mappings_tuple_value_change(self, style):
        check_list_editor_factory = CheckListEditor(
            values=[(1, "one"), (2, "two")],
            format_func=lambda t: t[1].upper(),
        )
        formatted_view = View(
            UItem(
                "value",
                editor=check_list_editor_factory,
                style=style,
            )
        )
        model = ListModel()

        with reraise_exceptions(), \
                self.setup_ui(model, formatted_view) as editor:

            # FIXME issue enthought/traitsui#841
            with self.assertRaises(AssertionError):
                self.assertEqual(editor.names, ["ONE", "TWO"])
            self.assertEqual(editor.names, ["one", "two"])

            check_list_editor_factory.values = [(2, "two"), (1, "one")]

            # FIXME issue enthought/traitsui#841
            with self.assertRaises(AssertionError):
                self.assertEqual(editor.names, ["TWO", "ONE"])
            self.assertEqual(editor.names, ["two", "one"])

    def check_checklist_mappings_name_change(self, style):
        class ListModel(HasTraits):
            value = List()
            possible_values = List(["one", "two"])

        check_list_editor_factory = CheckListEditor(
            name="object.possible_values",
            format_func=lambda v: v.upper(),
        )
        formatted_view = View(
            UItem(
                "value",
                editor=check_list_editor_factory,
                style=style,
            )
        )
        model = ListModel()

        with reraise_exceptions(), \
                self.setup_ui(model, formatted_view) as editor:

            self.assertEqual(editor.names, ["ONE", "TWO"])

            model.possible_values = ["two", "one"]

            self.assertEqual(editor.names, ["TWO", "ONE"])

    def check_checklist_mappings_tuple_name_change(self, style):
        class ListModel(HasTraits):
            value = List()
            possible_values = List([(1, "one"), (2, "two")])

        check_list_editor_factory = CheckListEditor(
            name="object.possible_values",
            format_func=lambda t: t[1].upper(),
        )
        formatted_view = View(
            UItem(
                "value",
                editor=check_list_editor_factory,
                style=style,
            )
        )
        model = ListModel()

        with reraise_exceptions(), \
                self.setup_ui(model, formatted_view) as editor:

            # FIXME issue enthought/traitsui#841
            with self.assertRaises(AssertionError):
                self.assertEqual(editor.names, ["ONE", "TWO"])
            self.assertEqual(editor.names, ["one", "two"])

            model.possible_values = [(2, "two"), (1, "one")]

            # FIXME issue enthought/traitsui#841
            with self.assertRaises(AssertionError):
                self.assertEqual(editor.names, ["TWO", "ONE"])
            self.assertEqual(editor.names, ["two", "one"])

    def check_checklist_values_change_after_ui_dispose(self, style):
        # Check the available values can change after the ui is closed

        class ListModel(HasTraits):
            value = List()
            possible_values = List(["one", "two"])

        check_list_editor_factory = CheckListEditor(
            name="object.possible_values",
        )
        view = View(
            UItem(
                "value",
                editor=check_list_editor_factory,
                style=style,
            )
        )
        model = ListModel()

        with reraise_exceptions():
            with create_ui(model, dict(view=view)):
                pass

            model.possible_values = ["two", "one"]

    def test_simple_editor_mapping_values(self):
        self.check_checklist_mappings_value_change("simple")

    def test_simple_editor_mapping_values_tuple(self):
        self.check_checklist_mappings_tuple_value_change("simple")

    def test_simple_editor_mapping_name(self):
        self.check_checklist_mappings_name_change("simple")

    def test_simple_editor_mapping_name_tuple(self):
        self.check_checklist_mappings_tuple_name_change("simple")

    def test_custom_editor_mapping_values(self):
        # FIXME issue enthought/traitsui#842
        if is_wx():
            import wx

            with self.assertRaises(wx._core.wxAssertionError):
                self.check_checklist_mappings_value_change("custom")
        else:
            self.check_checklist_mappings_value_change("custom")

    def test_custom_editor_mapping_values_tuple(self):
        # FIXME issue enthought/traitsui#842
        if is_wx():
            import wx

            with self.assertRaises(wx._core.wxAssertionError):
                self.check_checklist_mappings_tuple_value_change("custom")
        else:
            self.check_checklist_mappings_tuple_value_change("custom")

    def test_custom_editor_mapping_name(self):
        # FIXME issue enthought/traitsui#842
        if is_wx():
            import wx

            with self.assertRaises(wx._core.wxAssertionError):
                self.check_checklist_mappings_name_change("custom")
        else:
            self.check_checklist_mappings_name_change("custom")

    def test_custom_editor_mapping_name_tuple(self):
        # FIXME issue enthought/traitsui#842
        if is_wx():
            import wx

            with self.assertRaises(wx._core.wxAssertionError):
                self.check_checklist_mappings_tuple_name_change("custom")
        else:
            self.check_checklist_mappings_tuple_name_change("custom")

    def test_simple_editor_checklist_values_change_dispose(self):
        self.check_checklist_values_change_after_ui_dispose("simple")

    def test_custom_editor_checklist_values_change_dispose(self):
        self.check_checklist_values_change_after_ui_dispose("custom")


@requires_toolkit([ToolkitName.qt, ToolkitName.wx])
class TestSimpleCheckListEditor(BaseTestMixin, unittest.TestCase):

    def setUp(self):
        BaseTestMixin.setUp(self)

    def tearDown(self):
        BaseTestMixin.tearDown(self)

    @contextlib.contextmanager
    def setup_gui(self, model, view):
        with create_ui(model, dict(view=view)) as ui:
            process_cascade_events()
            editor = ui.get_editors("value")[0]
            yield editor

    def test_simple_check_list_editor_text(self):
        list_edit = ListModel(value=["one"])

        with reraise_exceptions(), \
                self.setup_gui(list_edit, get_view("simple")) as editor:

            self.assertEqual(get_combobox_text(editor.control), "One")

            list_edit.value = ["two"]
            process_cascade_events()

            self.assertEqual(get_combobox_text(editor.control), "Two")

    def test_simple_check_list_editor_text_mapped(self):
        view = get_mapped_view("simple")
        list_edit = ListModel(value=[1])

        with reraise_exceptions(), \
                self.setup_gui(list_edit, view) as editor:

            # FIXME issue enthought/traitsui#841
            with self.assertRaises(AssertionError):
                self.assertEqual(get_combobox_text(editor.control), "One")
            self.assertEqual(get_combobox_text(editor.control), "one")

            list_edit.value = [2]
            process_cascade_events()

            # FIXME issue enthought/traitsui#841
            with self.assertRaises(AssertionError):
                self.assertEqual(get_combobox_text(editor.control), "Two")
            self.assertEqual(get_combobox_text(editor.control), "two")

    def test_simple_check_list_editor_index(self):
        list_edit = ListModel(value=["one"])

        with reraise_exceptions(), \
                self.setup_gui(list_edit, get_view("simple")) as editor:

            self.assertEqual(list_edit.value, ["one"])

            set_combobox_index(editor, 1)
            process_cascade_events()

            self.assertEqual(list_edit.value, ["two"])

            set_combobox_index(editor, 0)
            process_cascade_events()

            self.assertEqual(list_edit.value, ["one"])

    def test_simple_check_list_editor_invalid_current_values(self):
        list_edit = ListModel(value=[1, "two", "a", object(), "one"])

        with reraise_exceptions(), \
                self.setup_gui(list_edit, get_view("simple")):

            self.assertEqual(list_edit.value, ["two", "one"])

    def test_simple_check_list_editor_invalid_current_values_str(self):
        class StrModel(HasTraits):
            value = Str()

        str_edit = StrModel(value="alpha, \ttwo, beta,\n lambda, one")

        with reraise_exceptions(), \
                self.setup_gui(str_edit, get_view("simple")):

            self.assertEqual(str_edit.value, "two,one")


@requires_toolkit([ToolkitName.qt, ToolkitName.wx])
class TestCustomCheckListEditor(BaseTestMixin, unittest.TestCase):

    def setUp(self):
        BaseTestMixin.setUp(self)

    def tearDown(self):
        BaseTestMixin.tearDown(self)

    @contextlib.contextmanager
    def setup_gui(self, model, view):
        with create_ui(model, dict(view=view)) as ui:
            process_cascade_events()
            editor = ui.get_editors("value")[0]
            yield editor

    def test_custom_check_list_editor_button_update(self):
        list_edit = ListModel()

        with reraise_exceptions(), \
                self.setup_gui(list_edit, get_view("custom")) as editor:

            self.assertEqual(
                get_all_button_status(editor.control),
                [False, False, False, False]
            )

            list_edit.value = ["two", "four"]
            process_cascade_events()

            self.assertEqual(
                get_all_button_status(editor.control),
                [False, True, False, True]
            )

            list_edit.value = ["one", "four"]
            process_cascade_events()

            self.assertEqual(
                get_all_button_status(editor.control),
                [True, False, False, True]
            )

    def test_custom_check_list_editor_click(self):
        list_edit = ListModel()

        tester = UITester()
        with tester.create_ui(list_edit, dict(view=get_view("custom"))) as ui:
            self.assertEqual(list_edit.value, [])
            check_list = tester.find_by_name(ui, "value")
            item_1 = check_list.locate(Index(1))
            item_1.perform(MouseClick())
            self.assertEqual(list_edit.value, ["two"])
            item_1.perform(MouseClick())
            self.assertEqual(list_edit.value, [])

    def test_custom_check_list_editor_click_initial_value(self):
        list_edit = ListModel(value=["two"])

        tester = UITester()
        with tester.create_ui(list_edit, dict(view=get_view("custom"))) as ui:
            self.assertEqual(list_edit.value, ["two"])

            check_list = tester.find_by_name(ui, "value")
            item_1 = check_list.locate(Index(1))
            item_1.perform(MouseClick())

            self.assertEqual(list_edit.value, [])

    def test_custom_check_list_editor_invalid_current_values_str(self):
        class StrModel(HasTraits):
            value = Str()

        str_edit = StrModel(value="alpha, \ttwo, three,\n lambda, one")

        tester = UITester()
        with tester.create_ui(str_edit, dict(view=get_view("custom"))) as ui:

            self.assertEqual(str_edit.value, "two,three,one")

            check_list = tester.find_by_name(ui, "value")
            item_1 = check_list.locate(Index(1))
            item_1.perform(MouseClick())

            self.assertEqual(str_edit.value, "three,one")

    def test_custom_check_list_editor_grid_layout(self):
        for cols in range(1, 8):
            list_edit = ListModel()
            tester = UITester()
            view = get_view_custom_cols(cols=cols)
            with tester.create_ui(list_edit, dict(view=view)) as ui:
                self.assertEqual(list_edit.value, [])
                check_list = tester.find_by_name(ui, "value")
                item = check_list.locate(Index(6))
                item.perform(MouseClick())
                self.assertEqual(list_edit.value, ["seven"])
                item.perform(MouseClick())
                self.assertEqual(list_edit.value, [])


@requires_toolkit([ToolkitName.qt, ToolkitName.wx])
class TestTextCheckListEditor(BaseTestMixin, unittest.TestCase):

    def setUp(self):
        BaseTestMixin.setUp(self)

    def tearDown(self):
        BaseTestMixin.tearDown(self)

    @contextlib.contextmanager
    def setup_gui(self, model, view):
        with create_ui(model, dict(view=view)) as ui:

            process_cascade_events()
            editor = ui.get_editors("value")[0]
            yield editor

    def test_text_check_list_object_list(self):
        list_edit = ListModel()

        with reraise_exceptions(), \
                self.setup_gui(list_edit, get_view("text")) as editor:

            self.assertEqual(list_edit.value, [])

            set_text_in_line_edit(editor.control, "['one', 'two']")
            process_cascade_events()

            self.assertEqual(list_edit.value, ["one", "two"])

    def test_text_check_list_object_str(self):
        class StrModel(HasTraits):
            value = Str()

        str_edit = StrModel(value="three, four")

        with reraise_exceptions(), \
                self.setup_gui(str_edit, get_view("text")) as editor:

            self.assertEqual(str_edit.value, "three, four")

            set_text_in_line_edit(editor.control, "one, two")
            process_cascade_events()

            self.assertEqual(str_edit.value, "one, two")



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
Demonstrates a 'display only' video editor for Qt5+.

Please refer to the `VideoEditor API docs`_ for further information.

.. _VideoEditor API docs: https://docs.enthought.com/traitsui/api/traitsui.editors.video_editor.html#traitsui.editors.video_editor.VideoEditor
"""
import numpy as np
from PIL import Image
from pyface.qt.QtGui import QImage
from traits.api import (
    Bool, Button, Callable, Float, HasTraits, Range, Str, observe
)
from traitsui.api import ButtonEditor, ContextValue, HGroup, Item, UItem, View
from traitsui.editors.video_editor import MediaStatus, PlayerState, VideoEditor


def QImage_from_np(image):
    assert (np.max(image) <= 255)
    image8 = image.astype(np.uint8, order='C', casting='unsafe')
    height, width, colors = image8.shape
    bytesPerLine = 4 * width

    image = QImage(image8.data, width, height, bytesPerLine,
                   QImage.Format_RGB32)
    return image


def np_from_QImage(qimage):
    # Creates a numpy array from a pyqt(5) QImage object
    width, height = qimage.width(), qimage.height()
    channels = qimage.pixelFormat().channelCount()
    return np.array(
        qimage.bits().asarray(width * height * channels)
    ).reshape(height, width, channels).astype('u1')


def qimage_function(antialiasing=True):
    def antialias_func(image_func):
        """
        Turns an image function into a QImage function,
        bound within the viewing frames box.
        """
        def qimage_conv_func(image, box_dims):
            _np_image = image_func(np_from_QImage(image))
            pil_image = Image.fromarray(_np_image)
            if antialiasing:
                pil_image.thumbnail(box_dims, Image.ANTIALIAS)
            else:
                pil_image.thumbnail(box_dims)
            _np_image = np.array(pil_image)
            image = QImage_from_np(_np_image)
            return image, _np_image
        return qimage_conv_func
    return antialias_func


@qimage_function(antialiasing=False)
def test_image_func(image):
    return image.transpose(1, 0, 2)


class VideoEditorDemo(HasTraits):
    """Defines the main VideoEditor demo class."""

    #: The URL that holds the video data.
    video_url = Str()

    #: A button that plays/pauses
    play_pause_button = Button()

    button_label = Str('Play')

    state = PlayerState

    duration = Float

    position = Range(low=0.0, high='duration')

    error = Str

    status = MediaStatus

    buffer = Range(0, 100)

    muted = Bool(True)

    volume = Range(0.0, 100.0, value=100.0)

    playback_rate = Float(1.0)

    image_func = Callable()

    @observe('state')
    def _state_update(self, event):
        if event.new == 'stopped' or event.new == 'paused':
            self.button_label = 'Play'
        elif event.new == 'playing':
            self.button_label = 'Pause'

    @observe('play_pause_button')
    def _play_pause_button_click(self, event):
        if self.state == 'stopped' or self.state == 'paused':
            self.state = 'playing'
        else:
            self.state = 'paused'

    # Demo view
    traits_view = View(
        UItem('video_url'),
        UItem(
            'video_url',
            editor=VideoEditor(
                state=ContextValue('state'),
                position=ContextValue('position'),
                duration=ContextValue('duration'),
                video_error=ContextValue('error'),
                media_status=ContextValue('status'),
                buffer=ContextValue('buffer'),
                muted=ContextValue('muted'),
                volume=ContextValue('volume'),
                playback_rate=ContextValue('playback_rate'),
                image_func=ContextValue('image_func'),
                notify_interval=0.5,
            ),
        ),
        HGroup(
            UItem(
                'play_pause_button',
                editor=ButtonEditor(label_value='button_label'),
                enabled_when='not bool(error)',
                width=100,
            ),
            UItem('position'),
        ),
        HGroup(
            Item('playback_rate'),
            Item('muted', label='Mute'),
            Item('volume'),
        ),
        HGroup(
            Item('state', style='readonly'),
            Item('status', style='readonly'),
            Item('buffer', style='readonly'),
            Item('error', visible_when="bool(error)", style='readonly'),
        ),
        title='VideoEditor',
        buttons=['OK'],
        width=800,
        height=600,
        resizable=True,
    )


url = "https://commondatastorage.googleapis.com/gtv-videos-bucket/sample/ForBiggerEscapes.mp4"  # noqa: E501
# Create the demo:
demo = VideoEditorDemo(image_func=test_image_func, video_url=url)
# demo = VideoEditorDemo(video_url=url)

# Run the demo (if invoked from the command line):
if __name__ == '__main__':
    demo.configure_traits()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import contextlib
import unittest

from traits.api import Enum, HasTraits, Int, List
from traitsui.api import EnumEditor, UItem, View
from traitsui.tests._tools import (
    BaseTestMixin,
    create_ui,
    is_qt,
    is_wx,
    requires_toolkit,
    reraise_exceptions,
    ToolkitName,
)
from traitsui.testing.api import (
    Disabled,
    DisplayedText,
    Index,
    IsEnabled,
    KeyClick,
    KeySequence,
    MouseClick,
    SelectedText,
    UITester
)


class EnumModel(HasTraits):

    value = Enum("one", "two", "three", "four")


class EnumModelWithNone(HasTraits):

    value = Enum(None, "one", "two", "three", "four")


def get_view(style):
    return View(UItem("value", style=style), resizable=True)


def get_evaluate_view(style, auto_set=True, mode="radio"):
    return View(
        UItem(
            "value",
            editor=EnumEditor(
                evaluate=True,
                values=["one", "two", "three", "four"],
                auto_set=auto_set,
                mode=mode,
            ),
            style=style,
        ),
        resizable=True,
    )


def get_radio_view(cols=1):
    return View(
        UItem(
            "value",
            editor=EnumEditor(
                values=["one", "two", "three", "four"],
                cols=cols,
                mode="radio",
            ),
            style="custom",
        ),
        resizable=True,
    )


@requires_toolkit([ToolkitName.qt, ToolkitName.wx])
class TestEnumEditorMapping(BaseTestMixin, unittest.TestCase):

    def setUp(self):
        BaseTestMixin.setUp(self)

    def tearDown(self):
        BaseTestMixin.tearDown(self)

    @contextlib.contextmanager
    def setup_ui(self, model, view):
        with create_ui(model, dict(view=view)) as ui:
            yield ui.get_editors("value")[0]

    def check_enum_mappings_value_change(self, style, mode):
        class IntEnumModel(HasTraits):
            value = Int()

        enum_editor_factory = EnumEditor(
            values=[0, 1],
            format_func=lambda v: str(bool(v)).upper(),
            mode=mode
        )
        formatted_view = View(
            UItem(
                "value",
                editor=enum_editor_factory,
                style=style,
            )
        )

        with reraise_exceptions(), \
                self.setup_ui(IntEnumModel(), formatted_view) as editor:

            self.assertEqual(editor.names, ["FALSE", "TRUE"])
            self.assertEqual(editor.mapping, {"FALSE": 0, "TRUE": 1})
            self.assertEqual(
                editor.inverse_mapping, {0: "FALSE", 1: "TRUE"}
            )

            enum_editor_factory.values = [1, 0]

            self.assertEqual(editor.names, ["TRUE", "FALSE"])
            self.assertEqual(editor.mapping, {"TRUE": 1, "FALSE": 0})
            self.assertEqual(
                editor.inverse_mapping, {1: "TRUE", 0: "FALSE"}
            )

    def check_enum_mappings_name_change(self, style, mode):
        class IntEnumModel(HasTraits):
            value = Int()
            possible_values = List([0, 1])

        formatted_view = View(
            UItem(
                'value',
                editor=EnumEditor(
                    name="object.possible_values",
                    format_func=lambda v: str(bool(v)).upper(),
                    mode=mode
                ),
                style=style,
            )
        )
        model = IntEnumModel()

        with reraise_exceptions(), \
                self.setup_ui(model, formatted_view) as editor:

            self.assertEqual(editor.names, ["FALSE", "TRUE"])
            self.assertEqual(editor.mapping, {"FALSE": 0, "TRUE": 1})
            self.assertEqual(
                editor.inverse_mapping, {0: "FALSE", 1: "TRUE"}
            )

            model.possible_values = [1, 0]

            self.assertEqual(editor.names, ["TRUE", "FALSE"])
            self.assertEqual(editor.mapping, {"TRUE": 1, "FALSE": 0})
            self.assertEqual(
                editor.inverse_mapping, {1: "TRUE", 0: "FALSE"}
            )

    def test_simple_editor_mapping_values(self):
        self.check_enum_mappings_value_change("simple", "radio")

    def test_simple_editor_mapping_name(self):
        self.check_enum_mappings_name_change("simple", "radio")

    def test_radio_editor_mapping_values(self):
        # FIXME issue enthought/traitsui#842
        if is_wx():
            import wx

            with self.assertRaises(wx._core.wxAssertionError):
                self.check_enum_mappings_value_change("custom", "radio")
        else:
            self.check_enum_mappings_value_change("custom", "radio")

    def test_radio_editor_mapping_name(self):
        # FIXME issue enthought/traitsui#842
        if is_wx():
            import wx

            with self.assertRaises(wx._core.wxAssertionError):
                self.check_enum_mappings_name_change("custom", "radio")
        else:
            self.check_enum_mappings_name_change("custom", "radio")

    def test_list_editor_mapping_values(self):
        self.check_enum_mappings_value_change("custom", "list")

    def test_list_editor_mapping_name(self):
        self.check_enum_mappings_name_change("custom", "list")


@requires_toolkit([ToolkitName.qt, ToolkitName.wx])
class TestSimpleEnumEditor(BaseTestMixin, unittest.TestCase):

    def setUp(self):
        BaseTestMixin.setUp(self)

    def tearDown(self):
        BaseTestMixin.tearDown(self)

    def check_enum_text_update(self, view):
        enum_edit = EnumModel()

        tester = UITester()
        with tester.create_ui(enum_edit, dict(view=view)) as ui:
            combobox = tester.find_by_name(ui, "value")
            displayed = combobox.inspect(DisplayedText())
            self.assertEqual(displayed, "one")

            combobox.locate(Index(1)).perform(MouseClick())
            displayed = combobox.inspect(DisplayedText())
            self.assertEqual(displayed, "two")

    def check_enum_object_update(self, view):
        enum_edit = EnumModel()

        tester = UITester()
        with tester.create_ui(enum_edit, dict(view=view)) as ui:

            self.assertEqual(enum_edit.value, "one")

            combobox = tester.find_by_name(ui, "value")
            for _ in range(3):
                combobox.perform(KeyClick("Backspace"))
            combobox.perform(KeySequence("two"))
            combobox.perform(KeyClick("Enter"))

            self.assertEqual(enum_edit.value, "two")

    def check_enum_index_update(self, view):
        enum_edit = EnumModel()
        tester = UITester()
        with tester.create_ui(enum_edit, dict(view=view)) as ui:

            self.assertEqual(enum_edit.value, "one")

            combobox = tester.find_by_name(ui, "value")
            combobox.locate(Index(1)).perform(MouseClick())

            self.assertEqual(enum_edit.value, "two")

    def check_enum_text_bad_update(self, view):
        enum_edit = EnumModel()

        tester = UITester()
        with tester.create_ui(enum_edit, dict(view=view)) as ui:

            self.assertEqual(enum_edit.value, "one")

            combobox = tester.find_by_name(ui, "value")
            for _ in range(3):
                combobox.perform(KeyClick("Backspace"))
            combobox.perform(KeyClick("H"))
            combobox.perform(KeyClick("Enter"))

            self.assertEqual(enum_edit.value, "one")

    def test_simple_enum_editor_text(self):
        self.check_enum_text_update(get_view("simple"))

    def test_simple_enum_editor_index(self):
        self.check_enum_index_update(get_view("simple"))

    def test_simple_evaluate_editor_text(self):
        self.check_enum_text_update(get_evaluate_view("simple"))

    def test_simple_evaluate_editor_index(self):
        self.check_enum_index_update(get_evaluate_view("simple"))

    def test_simple_evaluate_editor_bad_text(self):
        self.check_enum_text_bad_update(get_evaluate_view("simple"))

    def test_simple_evaluate_editor_object(self):
        self.check_enum_object_update(get_evaluate_view("simple"))

    def test_simple_evaluate_editor_object_no_auto_set(self):
        view = get_evaluate_view("simple", auto_set=False)
        enum_edit = EnumModel()

        tester = UITester()
        with tester.create_ui(enum_edit, dict(view=view)) as ui:
            self.assertEqual(enum_edit.value, "one")

            combobox = tester.find_by_name(ui, "value")
            for _ in range(3):
                combobox.perform(KeyClick("Backspace"))
            combobox.perform(KeySequence("two"))

            self.assertEqual(enum_edit.value, "one")
            combobox.perform(KeyClick("Enter"))
            self.assertEqual(enum_edit.value, "two")

    def test_simple_editor_resizable(self):
        # Smoke test for `qt4.enum_editor.SimpleEditor.set_size_policy`
        enum_edit = EnumModel()
        resizable_view = View(UItem("value", style="simple", resizable=True))

        tester = UITester()
        with tester.create_ui(enum_edit, dict(view=resizable_view)):
            pass

    def test_simple_editor_rebuild_editor_evaluate(self):
        # Smoke test for `wx.enum_editor.SimpleEditor.rebuild_editor`
        enum_editor_factory = EnumEditor(
            evaluate=True,
            values=["one", "two", "three", "four"],
        )
        view = View(UItem("value", editor=enum_editor_factory, style="simple"))

        tester = UITester()
        with tester.create_ui(EnumModel(), dict(view=view)):
            enum_editor_factory.values = ["one", "two", "three"]

    def test_simple_editor_disabled(self):
        enum_edit = EnumModel(value="two")
        view = View(
            UItem(
                "value",
                style="simple",
                enabled_when="value == 'one'",
                editor=EnumEditor(evaluate=True, values=["one", "two"]),
            ),
        )
        tester = UITester()
        with tester.create_ui(enum_edit, dict(view=view)) as ui:
            combobox = tester.find_by_name(ui, "value")
            with self.assertRaises(Disabled):
                combobox.perform(KeyClick("Enter"))
            with self.assertRaises(Disabled):
                combobox.perform(KeySequence("two"))
            self.assertFalse(combobox.inspect(IsEnabled()))


@requires_toolkit([ToolkitName.qt, ToolkitName.wx])
class TestRadioEnumEditor(BaseTestMixin, unittest.TestCase):

    def setUp(self):
        BaseTestMixin.setUp(self)

    def tearDown(self):
        BaseTestMixin.tearDown(self)

    def test_radio_enum_editor_update(self):
        enum_edit = EnumModel()
        tester = UITester()

        with tester.create_ui(enum_edit, dict(view=get_view("custom"))) as ui:
            # sanity check
            self.assertEqual(enum_edit.value, "one")
            radio_editor = tester.find_by_name(ui, "value")
            self.assertEqual(
                radio_editor.inspect(SelectedText()),
                "One",
            )

            enum_edit.value = "two"

            self.assertEqual(
                radio_editor.inspect(SelectedText()),
                "Two",
            )

    def test_radio_enum_editor_pick(self):
        for cols in range(1, 4):
            for row_major in [True, False]:
                enum_edit = EnumModel()
                tester = UITester()
                view = get_radio_view(cols=cols)
                with tester.create_ui(enum_edit, dict(view=view)) as ui:
                    # sanity check
                    self.assertEqual(enum_edit.value, "one")
                    radio_editor = tester.find_by_name(ui, "value")
                    if is_qt():
                        radio_editor._target.row_major = row_major
                        radio_editor._target.rebuild_editor()
                    item = radio_editor.locate(Index(3))
                    item.perform(MouseClick())
                    self.assertEqual(enum_edit.value, "four")

    def test_radio_enum_none_selected(self):
        enum_edit = EnumModelWithNone()
        tester = UITester()
        with tester.create_ui(enum_edit, dict(view=get_radio_view())) as ui:
            self.assertEqual(enum_edit.value, None)
            radio_editor = tester.find_by_name(ui, "value")
            displayed = radio_editor.inspect(SelectedText())
            self.assertEqual(displayed, None)


@requires_toolkit([ToolkitName.qt, ToolkitName.wx])
class TestListEnumEditor(BaseTestMixin, unittest.TestCase):

    def setUp(self):
        BaseTestMixin.setUp(self)

    def tearDown(self):
        BaseTestMixin.tearDown(self)

    def check_enum_text_update(self, view):
        enum_edit = EnumModel()

        tester = UITester()
        with tester.create_ui(enum_edit, dict(view=view)) as ui:

            list_editor = tester.find_by_name(ui, "value")
            displayed = list_editor.inspect(SelectedText())

            self.assertEqual(displayed, "one")

            list_editor.locate(Index(1)).perform(MouseClick())
            displayed = list_editor.inspect(SelectedText())
            self.assertEqual(displayed, "two")

    def check_enum_index_update(self, view):
        enum_edit = EnumModel()
        tester = UITester()
        with tester.create_ui(enum_edit, dict(view=view)) as ui:

            self.assertEqual(enum_edit.value, "one")

            list_editor = tester.find_by_name(ui, "value")
            list_editor.locate(Index(1)).perform(MouseClick())

            self.assertEqual(enum_edit.value, "two")

    def test_list_enum_editor_text(self):
        view = View(
            UItem(
                "value",
                editor=EnumEditor(
                    values=["one", "two", "three", "four"],
                    mode="list",
                ),
                style="custom",
            ),
            resizable=True,
        )
        self.check_enum_text_update(view)

    def test_list_enum_editor_index(self):
        view = View(
            UItem(
                "value",
                editor=EnumEditor(
                    values=["one", "two", "three", "four"],
                    mode="list",
                ),
                style="custom",
            ),
            resizable=True,
        )
        self.check_enum_index_update(view)

    def test_list_evaluate_editor_text(self):
        self.check_enum_text_update(get_evaluate_view("custom", mode="list"))

    def test_list_evaluate_editor_index(self):
        self.check_enum_index_update(get_evaluate_view("custom", mode="list"))

    def test_list_enum_none_selected(self):
        enum_edit = EnumModelWithNone()
        view = View(
            UItem(
                "value",
                editor=EnumEditor(
                    # Note that for the list style editor, in order to select
                    # None, it must be one of the displayed options
                    values=[None, "one", "two", "three", "four"],
                    mode="list",
                ),
                style="custom",
            ),
            resizable=True,
        )
        tester = UITester()
        with tester.create_ui(enum_edit, dict(view=view)) as ui:
            self.assertEqual(enum_edit.value, None)
            list_editor = tester.find_by_name(ui, "value")
            # As a result the displayed text is actually the string 'None'
            displayed = list_editor.inspect(SelectedText())
            self.assertEqual(displayed, 'None')



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
Implementation of a ButtonEditor demo plugin for Traits UI demo program.

This demo shows each of the two styles of the ButtonEditor.
(As of this writing, they are identical.)

Please refer to the `ButtonEditor API docs`_ for further information.

.. _ButtonEditor API docs: https://docs.enthought.com/traitsui/api/traitsui.editors.button_editor.html#traitsui.editors.button_editor.ButtonEditor
"""

from traits.api import HasTraits, Button
from traitsui.api import Item, View, Group, message


# -------------------------------------------------------------------------
#  Demo Class
# -------------------------------------------------------------------------

class ButtonEditorDemo(HasTraits):
    """ This class specifies the details of the ButtonEditor demo.
    """

    # To demonstrate any given Trait editor, an appropriate Trait is required.
    fire_event = Button('Click Me')

    def _fire_event_fired():
        message("Button clicked!")

    # ButtonEditor display
    # (Note that Text and ReadOnly versions are not applicable)
    event_group = Group(
        Item('fire_event', style='simple', label='Simple', id="simple"),
        Item('_'),
        Item('fire_event', style='custom', label='Custom', id="custom"),
        Item('_'),
        Item(label='[text style unavailable]'),
        Item('_'),
        Item(label='[readonly style unavailable]')
    )

    # Demo view
    traits_view = View(
        event_group,
        title='ButtonEditor',
        buttons=['OK'],
        width=250
    )


# Create the demo:
demo = ButtonEditorDemo()

# Run the demo (if invoked from the command line):
if __name__ == '__main__':
    demo.configure_traits()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Tests to exercise logic for layouts, e.g. VGroup, HGroup.
"""

import unittest

from pyface.toolkit import toolkit_object
from pyface.constant import OK
try:
    from pyface.qt import QtWebkit  # noqa: F401
    NO_WEBKIT_OR_WEBENGINE = False
except ImportError:
    try:
        from pyface.qt import QtWebEngine  # noqa: F401
        NO_WEBKIT_OR_WEBENGINE = False
    except ImportError:
        NO_WEBKIT_OR_WEBENGINE = True
from traits.api import HasTraits, Int
from traitsui.api import HelpButton, HGroup, Item, spring, VGroup, View
from traitsui.testing.api import MouseClick, UITester
from traitsui.tests._tools import (
    BaseTestMixin,
    create_ui,
    requires_toolkit,
    ToolkitName,
)

ModalDialogTester = toolkit_object(
    "util.modal_dialog_tester:ModalDialogTester"
)


class ObjectWithNumber(HasTraits):
    number1 = Int()
    number2 = Int()
    number3 = Int()


class HelpPanel(HasTraits):
    my_int = Int(2, help='this is the help for my int')

    def default_traits_view(self):
        view = View(
            Item(name="my_int"),
            title="HelpPanel",
            buttons=[HelpButton],
        )
        return view


@requires_toolkit([ToolkitName.qt, ToolkitName.wx])
class TestUIPanel(BaseTestMixin, unittest.TestCase):

    def setUp(self):
        BaseTestMixin.setUp(self)

    def tearDown(self):
        BaseTestMixin.tearDown(self)

    def test_grouped_layout_with_springy(self):
        # Regression test for enthought/traitsui#1066
        obj1 = ObjectWithNumber()
        view = View(
            HGroup(
                VGroup(
                    Item("number1"),
                ),
                VGroup(
                    Item("number2"),
                ),
                spring,
            )
        )
        # This should not fail.
        with create_ui(obj1, dict(view=view)):
            pass

    # Regression test for enthought/traitsui#1538
    @requires_toolkit([ToolkitName.qt])
    @unittest.skipIf(
        NO_WEBKIT_OR_WEBENGINE, "Tests require either QtWebKit or QtWebEngine"
    )
    def test_show_help(self):

        # This help window is not actually modal, when opened it will be the
        # active window not active modal widget
        class MyTester(ModalDialogTester):
            def get_dialog_widget(self):
                return self._qt_app.activeWindow()

        panel = HelpPanel()
        tester = UITester(auto_process_events=False)
        with tester.create_ui(panel) as ui:
            help_button = tester.find_by_id(ui, 'Help')

            def click_help():
                # should not fail
                help_button.perform(MouseClick())

            mdtester = MyTester(click_help)
            mdtester.open_and_run(lambda x: x.click_button(OK))



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Utility functions for handling exceptions from the GUI event loop.

In the test context, uncaught exceptions should cause tests to error.
"""

from contextlib import contextmanager
import logging
import sys
import traceback

_TRAITSUI_LOGGER = logging.getLogger("traitsui")


def _serialize_exception(exc_type, value, tb):
    """ Serialize exception and traceback for reporting.
    This is such that the stack frame is not prevented from being garbage
    collected.
    """
    return (
        str(exc_type),
        str(value),
        str("".join(traceback.format_exception(exc_type, value, tb)))
    )


@contextmanager
def reraise_exceptions(logger=_TRAITSUI_LOGGER):
    """ Context manager to capture all exceptions occurred in the context and
    then reraise a RuntimeError if there are any exceptions captured.

    Depending on the GUI toolkit backend, unexpected exceptions occurred in the
    GUI event loop may (1) cause fatal early exit of the test suite or (2) be
    printed to the console without causing the test to error. This context
    manager is intended for testing purpose such that unexpected exceptions
    will result in a test error.

    Parameters
    ----------
    logger : logging.Logger
        Logger to use for logging errors.
    """
    serialized_exceptions = []

    def excepthook(type, value, tb):
        serialized = _serialize_exception(type, value, tb)
        serialized_exceptions.append(serialized)
        logger.error(
            "Unexpected error captured in sys excepthook. \n%s",
            serialized[-1]
        )

    sys.excepthook = excepthook
    try:
        yield
    finally:
        sys.excepthook = sys.__excepthook__
        if serialized_exceptions:
            msg = "Uncaught exceptions found.\n"
            msg += "\n".join(
                "=== Exception (type: {}, value: {}) ===\n"
                "{}".format(*record)
                for record in serialized_exceptions
            )
            raise RuntimeError(msg)



# (C) Copyright 2009-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# ------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#
# ------------------------------------------------------------------------------

""" Defines the editor factory for multi-selection enumerations, for all traits
toolkit backends.
"""

from traits.api import Range

from traitsui.editor_factory import EditorWithListFactory


class CheckListEditor(EditorWithListFactory):
    """ Editor factory for checklists.
    """

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: Number of columns to use when the editor is displayed as a grid
    cols = Range(1, 20)


# This alias is deprecated and will be removed in TraitsUI 8.
ToolkitEditorFactory = CheckListEditor



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# wizard.py -- Example of a traits-based wizard UI

from traits.api import HasTraits, Str
from traits.etsconfig.api import ETSConfig
from traitsui.api import Item, View, VGroup


class Person(HasTraits):
    first_name = Str()
    last_name = Str()

    company = Str()
    position = Str()

    view = View(
        VGroup(Item("first_name"), Item("last_name")),
        VGroup(Item("company"), Item("position")),
    )


person = Person(
    first_name='Postman',
    last_name='Pat',
    company="Enthought",
    position="Software Developer",
)


if ETSConfig.toolkit == "wx":
    # Wizard window is currently only available for wx backend.
    person.configure_traits(kind='wizard')



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""Traits UI 'display only' video editor."""
from pyface.qt.QtCore import QPoint, Qt, QUrl, Signal
from pyface.qt.QtGui import QImage, QPainter, QPalette, QSizePolicy
from pyface.qt.QtMultimedia import (QAbstractVideoBuffer,
                                    QAbstractVideoSurface, QAudio,
                                    QMediaContent, QMediaPlayer, QVideoFrame)
from pyface.qt.QtMultimediaWidgets import QVideoWidget
from traits.api import Bool, Callable, Float, Instance, Range, Str, observe
from traitsui.editors.video_editor import AspectRatio, MediaStatus, PlayerState

from .editor import Editor

#: Map from ApectRatio enum values to Qt aspect ratio behaviours.
aspect_ratio_map = {
    'ignore': Qt.IgnoreAspectRatio,
    'keep': Qt.KeepAspectRatio,
    'expand': Qt.KeepAspectRatioByExpanding,
}

#: Map from PlayerState enum values to QMediaPlayer states.
state_map = {
    'stopped': QMediaPlayer.StoppedState,
    'playing': QMediaPlayer.PlayingState,
    'paused': QMediaPlayer.PausedState,
}

#: Map from QMediaPlayer states to PlayerState enum values.
reversed_state_map = {value: key for key, value in state_map.items()}

#: Map from QMediaPlayer media status values to MediaStatus enum values.
media_status_map = {
    QMediaPlayer.UnknownMediaStatus: 'unknown',
    QMediaPlayer.NoMedia: 'no_media',
    QMediaPlayer.LoadingMedia: 'loading',
    QMediaPlayer.LoadedMedia: 'loaded',
    QMediaPlayer.StalledMedia: 'stalled',
    QMediaPlayer.BufferingMedia: 'buffering',
    QMediaPlayer.BufferedMedia: 'buffered',
    QMediaPlayer.EndOfMedia: 'end',
    QMediaPlayer.InvalidMedia: 'invalid',
}


class ImageWidget(QVideoWidget):
    """ Paints a QImage to the window body. """

    def __init__(self, parent=None, image_func=None):
        import numpy as np
        super().__init__(parent)
        self.image = QImage()
        self._np_image = np.zeros(shape=(0, 0, 4))
        self.painter = None
        self.resizeEvent(None)
        if image_func is None:
            def I_fun(image, bbox):
                # Don't bother with creating an ndarray version
                return image, self._np_image
            self.image_func = I_fun
        else:
            self.image_func = image_func

    def resizeEvent(self, event):
        s = self.size()
        self.width = s.width()
        self.height = s.height()

    def setImage(self, image):
        self.image, self._np_image = self.image_func(
            image, (self.width, self.height)
        )
        self.update()

    def paintEvent(self, event):
        super().paintEvent(event)
        if self.painter is None:
            self.painter = QPainter()
        self.painter.begin(self)
        if self.image:
            self.painter.drawImage(QPoint(0, 0), self.image)
        self.painter.end()


class VideoSurface(QAbstractVideoSurface):

    frameAvailable = Signal(['QImage'])

    def __init__(self, widget=None):
        super().__init__()
        self.widget = widget

    def supportedPixelFormats(self, handleType):
        return [QVideoFrame.Format_RGB32]

    def present(self, frame):
        cloned_frame = QVideoFrame(frame)
        cloned_frame.map(QAbstractVideoBuffer.ReadOnly)
        image = QImage(
            cloned_frame.bits(),
            cloned_frame.width(),
            cloned_frame.height(),
            cloned_frame.bytesPerLine(),
            QVideoFrame.imageFormatFromPixelFormat(cloned_frame.pixelFormat())
        )
        self.frameAvailable.emit(image)
        return True


class VideoEditor(Editor):
    """Traits UI 'display only' video editor.

    This editor uses the Qt QMultimedia machinery to display video streams
    to the screen.  Rather than being self-contained, the editor only concerns
    itself with displaying the video, and provides traits that control
    behaviour and provide internal state of the control during playback.
    """

    #: Does the drawing onto the image plane
    control = Instance(QVideoWidget)

    #: Handles the image pulling so the frames can be processed.
    surface = Instance(QAbstractVideoSurface)

    #: The QMediaObject that holds the connection to the video stream.
    media_content = Instance(QMediaContent)

    #: The QMediaPlayer that controls playback of the video stream.
    media_player = Instance(QMediaPlayer)

    #: The aspect ratio of the video editor.
    aspect_ratio = AspectRatio()

    #: The current state of the player, synchronized to the trait named
    #: by factory.state.
    state = PlayerState()

    #: The current playback position of the player, synchronized to the trait
    #: named by factory.position.
    position = Float()

    #: The total duration of the current video, synchronized to the trait
    #: named by factory.duration.
    duration = Float()

    #: The media player playback status (loading, buffering, etc.),
    #: synchronized to the trait named by factory.media_status.
    media_status = MediaStatus()

    #: The percentage of the buffer currently filled, synchronized to the trait
    #: named by factory.buffer.
    buffer = Range(0, 100)

    #: A string holding the video error state, or "" if no error. Synchronized
    #: to the trait named by factory.video_error.
    video_error = Str()

    #: Whether the audio is muted or not, synchronized to the trait named by
    #: factory.muted.
    muted = Bool(False)

    #: The playback volume on a logarithmic scale (perceptually linear),
    #: synchronized to the trait named by factory.volume.
    volume = Range(0.0, 100.0)

    #: The playback rate.  Negative values rewind the video.
    #: Synchronized to the trait named by factory.playback_rate.
    playback_rate = Float(1.0)

    #: Function to apply to the image. Takes ref to new frame and a size tuple.
    #: Synchronized to the trait named by factory.image_func.
    image_func = Callable()

    #: The change in position required for an update to be emitted.
    #: Synchronized to the trait named by factory.notify_interval.
    notify_interval = Float(1.0)

    def update_to_regular(self):
        if self.surface is not None:
            self.surface.frameAvailable.disconnect(self.control.setImage)
            self.surface = None

        self.control = QVideoWidget()
        self.control.setSizePolicy(
            QSizePolicy.Expanding, QSizePolicy.Expanding
        )
        self.control.setBackgroundRole(QPalette.Window)
        self.media_player.setVideoOutput(self.control)

    def update_to_functional(self):
        self.control = ImageWidget(image_func=self.image_func)
        self.control.setSizePolicy(
            QSizePolicy.Expanding, QSizePolicy.Expanding
        )
        self.control.setBackgroundRole(QPalette.Window)

        self.surface = VideoSurface(widget=self.control)
        self.surface.frameAvailable.connect(self.control.setImage)

        self.media_player.setVideoOutput(self.surface)

    # ------------------------------------------------------------------------
    # Editor interface
    # ------------------------------------------------------------------------

    def init(self, parent):
        """Initialize the editor by creating the underlying toolkit widget.

        Parameters
        ----------
        parent : QWidget or None
            The parent widget for this widget.
        """
        self.control = QVideoWidget()
        self.control.setSizePolicy(
            QSizePolicy.Expanding, QSizePolicy.Expanding
        )
        self.control.setBackgroundRole(QPalette.Window)

        self.media_player = QMediaPlayer(None, QMediaPlayer.VideoSurface)
        self._set_video_url()
        self.media_player.setVideoOutput(self.control)
        self.media_player.setMuted(self.muted)
        self._update_state()
        self._update_aspect_ratio()
        self._update_muted()
        self._update_volume()
        self._update_playback_rate()
        self._update_notify_interval()

        self._connect_signals()

        self.set_tooltip()

    def dispose(self):
        self._disconnect_signals()
        if self.media_player is not None:
            # Avoid a segfault if the media player is currently playing
            self.media_player.setVideoOutput(None)

        super().dispose()

    def update_editor(self):
        """Update the editor when the object trait changes externally."""
        self._set_video_url()

    # ------------------------------------------------------------------------
    # Private interface
    # ------------------------------------------------------------------------

    def _connect_signals(self):
        if self.media_player is not None:
            self.media_player.stateChanged.connect(self._state_changed_emitted)
            self.media_player.positionChanged.connect(
                self._position_changed_emitted)
            self.media_player.durationChanged.connect(
                self._duration_changed_emitted)
            self.media_player.error.connect(self._error_emitted)
            self.media_player.mediaStatusChanged.connect(
                self._media_status_changed_emitted)
            self.media_player.bufferStatusChanged.connect(
                self._buffer_status_changed_emitted)
            self.media_player.notifyIntervalChanged.connect(
                self._notify_interval_changed_emitted)

    def _disconnect_signals(self):
        if self.media_player is not None:
            self.media_player.stateChanged.disconnect(
                self._state_changed_emitted)
            self.media_player.positionChanged.disconnect(
                self._position_changed_emitted)
            self.media_player.durationChanged.disconnect(
                self._duration_changed_emitted)
            self.media_player.error.disconnect(self._error_emitted)
            self.media_player.mediaStatusChanged.disconnect(
                self._media_status_changed_emitted)
            self.media_player.bufferStatusChanged.disconnect(
                self._buffer_status_changed_emitted)
            self.media_player.notifyIntervalChanged.disconnect(
                self._notify_interval_changed_emitted)

    def _set_video_url(self):
        qurl = QUrl.fromUserInput(self.value)
        if qurl.isValid():
            self.media_content = QMediaContent(qurl)
            self.control.updateGeometry()
        else:
            self.media_content = QMediaContent(None)
            self.control.updateGeometry()

    # Signal handlers -------------------------------------------------------

    def _state_changed_emitted(self, state):
        self.state = reversed_state_map[state]

    def _position_changed_emitted(self, position):
        # Avoid telling Qt about the new position in `_position_changed`
        with self.updating_value():
            self.position = position / 1000.0

    def _duration_changed_emitted(self, duration):
        self.duration = duration / 1000.0

    def _error_emitted(self, error):
        if error != QMediaPlayer.NoError:
            self.video_error = self.media_player.errorString()
        else:
            self.video_error = ''

    def _media_status_changed_emitted(self, error):
        self.media_status = media_status_map[self.media_player.mediaStatus()]

    def _buffer_status_changed_emitted(self, error):
        self.buffer = self.media_player.bufferStatus()

    def _notify_interval_changed_emitted(self, interval):
        self.notify_interval = interval / 1000.0

    # Trait change handlers -------------------------------------------------

    @observe('aspect_ratio')
    def _aspect_ratio_observer(self, event):
        if self.control is not None:
            self._update_aspect_ratio()

    @observe('image_func')
    def _image_func_observer(self, event):
        if self.image_func is None:
            self.update_to_regular()
        else:
            self.update_to_functional()

    @observe('media_content')
    def _media_content_observer(self, event):
        self.video_error = ''
        if self.media_player is not None:
            self.media_player.setMedia(self.media_content)

    @observe('muted')
    def _muted_observer(self, event):
        if self.media_player is not None:
            self._update_muted()

    @observe('playback_rate')
    def _playback_rate_observer(self, event):
        if self.media_player is not None:
            self._update_playback_rate()

    @observe('position')
    def _position_observer(self, event):
        if self.media_player is not None and not self.updating:
            # position is given in ms
            self.media_player.setPosition(int(self.position * 1000))

    @observe('state')
    def _state_observer(self, event):
        if self.media_player is not None:
            self._update_state()

    @observe('volume')
    def _volume_observer(self, event):
        if self.media_player is not None:
            self._update_volume()

    @observe('notify_interval')
    def _notify_interval_observer(self, event):
        if self.media_player is not None:
            self._update_notify_interval()

    # MediaPlayer management ------------------------------------------------

    def _update_aspect_ratio(self):
        self.control.setAspectRatioMode(aspect_ratio_map[self.aspect_ratio])

    def _update_muted(self):
        self.media_player.setMuted(self.muted)

    def _update_playback_rate(self):
        self.media_player.setPlaybackRate(self.playback_rate)

    def _update_state(self):
        if self.state == 'stopped':
            self.media_player.stop()
            self.control.repaint()
        elif self.state == 'playing':
            # XXX forcing a resize so video is scaled correctly on MacOS
            s = self.control.size()
            w = s.width()
            h = s.height()
            self.media_player.play()
            self.control.resize(w+1, h+1)
            self.control.resize(w, h)
        elif self.state == 'paused':
            self.media_player.pause()

    def _update_volume(self):
        linear_volume = QAudio.convertVolume(
            self.volume/100.0,
            QAudio.LogarithmicVolumeScale,
            QAudio.LinearVolumeScale,
        )
        self.media_player.setVolume(int(linear_volume * 100))

    def _update_notify_interval(self):
        # interval is given in ms
        interval = int(self.notify_interval * 1000)
        self.media_player.setNotifyInterval(interval)



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
Enum editor

The Enum editor provides a simple way for the user to choose one item from
a list of known values (normally strings or numbers).

An Enum trait can take any value from a specified list of values. These values
are typically strings, integers, or floats, but can also be None or hashable
tuples.

An Enum can be displayed / edited in one of five styles:

* 'simple' displays a drop-down list of allowed values
* 'custom' by default, displays one or more columns of radio buttons (only
  one of which is selected at a time).
* 'custom' in 'list' mode (see source code below), displays a list of all the
  allowed values at once.
* 'readonly' displays the current value as non-editable text.
* 'text' displays the current value as text. You can also edit this text,
  but your text must be in the list of allowed values.

Please refer to the `EnumEditor API docs`_ for further information.

.. _EnumEditor API docs: https://docs.enthought.com/traitsui/api/traitsui.editors.enum_editor.html#traitsui.editors.enum_editor.EnumEditor
"""

from traits.api import HasTraits, Enum

from traitsui.api import Item, Group, View, EnumEditor


class EnumEditorDemo(HasTraits):
    """ Defines the main EnumEditor demo class. """

    # Define an Enum trait to view.
    name_list = Enum('A-495', 'A-498', 'R-1226', 'TS-17', 'TS-18',
                     'Foo', 12345, (11, 7), None)

    # Items are used to define the display, one Item per editor style:
    enum_group = Group(
        Item('name_list', style='simple', label='Simple', id="simple"),
        Item('_'),

        # The simple style also supports text entry:
        Item(
            'name_list',
            style='simple',
            label='Simple (text entry)',
            editor=EnumEditor(
                values=name_list,
                completion_mode='popup',
                evaluate=True
            ),
            id="simple_text"
        ),
        Item('_'),

        # The custom style defaults to radio button mode:
        Item('name_list', style='custom', label='Custom radio', id="radio"),
        Item('_'),

        # The custom style can also display in list mode, with extra work:
        Item(
            'name_list',
            style='custom',
            label='Custom list',
            editor=EnumEditor(values=name_list, mode='list'),
            id="list"
        ),
        Item('_'),

        Item('name_list', style='text', label='Text', id="text"),
        Item('_'),

        Item('name_list', style='readonly', label='ReadOnly', id="readonly")
    )

    # Demo view:
    traits_view = View(
        enum_group,
        title='EnumEditor',
        buttons=['OK'],
        resizable=True
    )


# Create the demo:
demo = EnumEditorDemo()

# Run the demo (if invoked from the command line):
if __name__ == '__main__':
    demo.configure_traits()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
import unittest

from traits.api import HasTraits

from traitsui.toolkit_traits import RGBColor


class HasRGBColor(HasTraits):
    color = RGBColor()


class TestRGBColor(unittest.TestCase):

    # regression test for enthought/traitsui#1531
    def test_hex_converion(self):
        has_rgb_color = HasRGBColor()
        has_rgb_color.color = 0x000000
        self.assertEqual(has_rgb_color.color, (0.0, 0.0, 0.0))



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
**WARNING**

  This demo might not work as expected and some documented features might be
  missing.

-------------------------------------------------------------------------------

This demonstrates using the TraitsUI file dialog with a file dialog extension,
in this case, the **TextInfo** extension, which displays (if possible) the
contents of the currently selected file in a read-only text editor so the user
can quickly verify they are opening the correct file before leaving the file
dialog.

For more information about why you would want to use the TraitsUI file dialog
over the standard OS file dialog, select the **File Open** demo. For a
demonstration of writing a custom file dialog extension, select the
**File Open with Custom Extension** demo.

This example also shows setting a file name filter which only allows Python
source (i.e. ``*.py``) files to be viewed and selected.
"""
# Issue related to the demo warning: enthought/traitsui#953

from traits.api import HasTraits, File, Button

from traitsui.api import View, HGroup, Item

from traitsui.file_dialog import open_file, TextInfo


# -- FileDialogDemo Class -------------------------------------------------

# Demo specific file dialog id:
demo_id = 'traitsui.demo.standard_editors.file_dialog.text_info'


class FileDialogDemo(HasTraits):

    # The name of the selected file:
    file_name = File()

    # The button used to display the file dialog:
    open = Button('Open...')

    # -- Traits View Definitions ----------------------------------------------

    traits_view = View(
        HGroup(
            Item('open', show_label=False),
            '_',
            Item('file_name', style='readonly', springy=True)
        ),
        width=0.5
    )

    # -- Traits Event Handlers ------------------------------------------------

    def _open_changed(self):
        """ Handles the user clicking the 'Open...' button.
        """
        file_name = open_file(extensions=TextInfo(),
                              filter='Python file (*.py)|*.py',
                              id=demo_id)
        if file_name != '':
            self.file_name = file_name


# Create the demo:
demo = FileDialogDemo()

# Run the demo (if invoked from the command line):
if __name__ == '__main__':
    demo.configure_traits()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
**WARNING**

  This demo might not work as expected and some documented features might be
  missing.

-------------------------------------------------------------------------------

This demonstrates using the TraitsUI file dialog with a custom written file
dialog extension, in this case an extension called **LineCountInfo**, which
displays the number of text lines in the currently selected file.

For more information about why you would want to use the TraitsUI file dialog
over the standard OS file dialog, select the **File Open** demo.
"""
# Issue related to the demo warning: enthought/traitsui#953

from os.path import getsize

from traits.api import HasTraits, File, Button, Property, cached_property

from traitsui.api import View, VGroup, HGroup, Item

from traitsui.file_dialog import open_file, MFileDialogModel


# -- LineCountInfo Class --------------------------------------------------
class LineCountInfo(MFileDialogModel):
    """ Defines a file dialog extension that displays the number of text lines
        in the currently selected file.
    """

    # The number of text lines in the currently selected file:
    lines = Property(observe='file_name')

    # -- Traits View Definitions ----------------------------------------------

    traits_view = View(
        VGroup(
            Item('lines', style='readonly'),
            label='Line Count Info',
            show_border=True
        )
    )

    # -- Property Implementations ---------------------------------------------

    @cached_property
    def _get_lines(self):
        try:
            if getsize(self.file_name) > 10000000:
                return 'File too big...'

            with open(self.file_name, 'r', encoding='utf8') as fh:
                data = fh.read()
        except OSError:
            return ''

        if (data.find('\x00') >= 0) or (data.find('\xFF') >= 0):
            return 'File contains binary data...'

        return ('{:n} lines'.format(len(data.splitlines())))


# -- FileDialogDemo Class -------------------------------------------------

# Demo specific file dialig id:
demo_id = 'traitsui.demo.standard_editors.file_dialog.line_count_info'


class FileDialogDemo(HasTraits):

    # The name of the selected file:
    file_name = File()

    # The button used to display the file dialog:
    open = Button('Open...')

    # -- Traits View Definitions ----------------------------------------------

    traits_view = View(
        HGroup(
            Item('open', show_label=False),
            '_',
            Item('file_name', style='readonly', springy=True)
        ),
        width=0.5
    )

    # -- Traits Event Handlers ------------------------------------------------

    def _open_changed(self):
        """ Handles the user clicking the 'Open...' button.
        """
        file_name = open_file(extensions=LineCountInfo(), id=demo_id)
        if file_name != '':
            self.file_name = file_name


# Create the demo:
demo = FileDialogDemo()

# Run the demo (if invoked from the command line):
if __name__ == '__main__':
    demo.configure_traits()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" This module provides a getter to obtain an AbstractTargetRegistry that can
support testing any target whose 'control' attribute refers to a QWidget.
"""
from pyface.qt import QtGui

from traitsui.testing.tester._dynamic_target_registry import (
    DynamicTargetRegistry,
)
from traitsui.testing.tester.query import IsEnabled, IsVisible


def _handle_is_enabled(wrapper, interaction):
    """ Return true if the target's control is enabled.

    Parameters
    ----------
    wrapper : UIWrapper
        Wrapper on which the target's control should be a QWidget
    interaction : IsEnabled
        Not currently used.
    """
    return wrapper._target.control.isEnabled()


def _handle_is_visible(wrapper, interaction):
    """ Return true if the target's control is visible.

    Parameters
    ----------
    wrapper : UIWrapper
        Wrapper on which the target's control should be a QWidget
    interaction : IsVisible
        Not currently used.
    """
    return wrapper._target.control.isVisible()


def _is_target_control_a_qt_widget(target):
    """ Return true if the target is accepted by the registry.

    Parameters
    ----------
    target : any
        Any UI target

    Returns
    -------
    is_accepted : bool
    """
    return (
        hasattr(target, "control")
        and isinstance(target.control, QtGui.QWidget)
    )


def get_widget_registry():
    """ Return a registry to support any target with an attribute 'control'
    that is an instance of QWidget.

    Returns
    -------
    registry : DynamicTargetRegistry
        A registry that can be used with UIWrapper
    """
    return DynamicTargetRegistry(
        can_support=_is_target_control_a_qt_widget,
        interaction_to_handler={
            IsEnabled: _handle_is_enabled,
            IsVisible: _handle_is_visible,
        }
    )



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
Displaying large NumPy arrays with TabularEditor

A demonstration of how the TabularEditor can be used to display (large) NumPy
arrays, in this case 100,000 random 3D points from a unit cube.

In addition to showing the coordinates of each point, it also displays the
index of each point in the array, as well as a red flag if the point lies
within 0.25 of the center of the cube.
"""

from numpy import sqrt
from numpy.random import random

from traits.api import HasTraits, Property, Array
from traitsui.api import View, Item, TabularAdapter, TabularEditor


# -- Tabular Adapter Definition -------------------------------------------
class ArrayAdapter(TabularAdapter):

    columns = [('i', 'index'), ('x', 0), ('y', 1), ('z', 2)]

    font = 'Courier 10'
    alignment = 'right'
    format = '%.4f'

    index_text = Property()
    index_image = Property()

    def _get_index_text(self):
        return str(self.row)

    def _get_index_image(self):
        x, y, z = self.item
        if sqrt((x - 0.5) ** 2 + (y - 0.5) ** 2 + (z - 0.5) ** 2) <= 0.25:
            return '@icons:red_ball'

        return None

# -- ShowArray Class Definition -------------------------------------------


class ShowArray(HasTraits):

    data = Array

    traits_view = View(
        Item(
            'data',
            show_label=False,
            editor=TabularEditor(
                adapter=ArrayAdapter(),
                auto_resize=True,
                # Do not allow any kind of editing of the array:
                editable=False,
                operations=[],
                drag_move=False
            )
        ),
        title='Array Viewer',
        width=0.3,
        height=0.8,
        resizable=True
    )


# Create the demo:
demo = ShowArray(data=random((100000, 3)))

# Run the demo (if invoked from the command line):
if __name__ == '__main__':
    demo.configure_traits()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
Test the layout of elements is consistent with the layout parameters.
"""

import unittest

from traits.api import Enum, HasTraits, Str

from traitsui.item import Item, UItem
from traitsui.view import View
from traitsui.group import HGroup, VGroup
from traitsui.testing.api import UITester
from traitsui.tests._tools import (
    BaseTestMixin,
    create_ui,
    requires_toolkit,
    reraise_exceptions,
    ToolkitName,
)


_DIALOG_WIDTH = 500
_DIALOG_HEIGHT = 500
_TXT_WIDTH = 100


class MultipleTrait(HasTraits):
    """ An object with multiple traits to test layout and alignments."""

    txt1 = Str("text1")
    txt2 = Str("text2")


class VResizeDialog(HasTraits):

    txt = Str("hallo")

    traits_view = View(
        VGroup(Item("txt", width=_TXT_WIDTH, resizable=True)),
        width=_DIALOG_WIDTH,
        height=_DIALOG_HEIGHT,
        resizable=True,
    )


class HResizeDialog(HasTraits):

    txt = Str("hallo")

    traits_view = View(
        HGroup(Item("txt", width=_TXT_WIDTH, resizable=True)),
        width=_DIALOG_WIDTH,
        height=_DIALOG_HEIGHT,
        resizable=True,
    )


class ObjectWithResizeReadonlyItem(HasTraits):
    resizable_readonly_item = Enum("first", "second")

    def default_traits_view(self):
        return View(
            VGroup(
                UItem(
                    "resizable_readonly_item",
                    resizable=True,
                    style="readonly",
                ),
            ),
            height=_DIALOG_HEIGHT,
            width=_DIALOG_WIDTH,
        )


class TestLayout(BaseTestMixin, unittest.TestCase):

    def setUp(self):
        BaseTestMixin.setUp(self)

    def tearDown(self):
        BaseTestMixin.tearDown(self)

    @requires_toolkit([ToolkitName.qt])
    def test_qt_resizable_in_vgroup(self):
        # Behavior: Item.resizable controls whether a component can resize
        # along the non-layout axis of its group. In a VGroup, resizing should
        # work only in the horizontal direction.

        with reraise_exceptions(), \
                create_ui(VResizeDialog()) as ui:
            editor, = ui.get_editors("txt")
            text = editor.control

            # horizontal size should be large
            self.assertGreater(text.width(), _DIALOG_WIDTH - 100)

            # vertical size should be unchanged
            self.assertLess(text.height(), 100)

    @requires_toolkit([ToolkitName.qt])
    def test_qt_resizable_in_hgroup(self):
        # Behavior: Item.resizable controls whether a component can resize
        # along the non-layout axis of its group. In a HGroup, resizing should
        # work only in the vertical direction.

        with reraise_exceptions(), \
                create_ui(HResizeDialog()) as ui:

            editor, = ui.get_editors("txt")
            text = editor.control

            # vertical size should be large
            self.assertGreater(text.height(), _DIALOG_HEIGHT - 100)

            # horizontal size should be unchanged
            # ??? maybe not: some elements (e.g., the text field) have
            # 'Expanding' as their default behavior
            # self.assertLess(text.width(), _TXT_WIDTH+100)

    # regression test for enthought/traitsui#1528
    @requires_toolkit([ToolkitName.qt])
    def test_qt_resizable_readonly_item(self):
        tester = UITester()
        with tester.create_ui(ObjectWithResizeReadonlyItem()) as ui:
            resizable_readonly_item = tester.find_by_name(
                ui, "resizable_readonly_item"
            )
            # for resizable item expansion should occur in horizontal but not
            # vertical direction
            self.assertLess(
                resizable_readonly_item._target.control.height(),
                _DIALOG_HEIGHT
            )
            self.assertEqual(
                resizable_readonly_item._target.control.width(),
                _DIALOG_WIDTH
            )


@requires_toolkit([ToolkitName.qt, ToolkitName.wx])
class TestOrientation(BaseTestMixin, unittest.TestCase):
    """ Toolkit-agnostic tests on the layout orientations."""

    def setUp(self):
        BaseTestMixin.setUp(self)

    def tearDown(self):
        BaseTestMixin.tearDown(self)

    def test_vertical_layout(self):
        view = View(
            VGroup(
                Item("txt1"),
                Item("txt2"),
            )
        )
        with reraise_exceptions(), \
                create_ui(MultipleTrait(), ui_kwargs=dict(view=view)):
            pass

    def test_horizontal_layout(self):
        # layout
        view = View(
            HGroup(
                Item("txt1"),
                Item("txt2"),
            )
        )
        with reraise_exceptions(), \
                create_ui(MultipleTrait(), ui_kwargs=dict(view=view)):
            pass


if __name__ == "__main__":
    # Execute from command line for manual testing
    vw = VResizeDialog()
    vw.configure_traits()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# mixed_styles.py -- Example of using editor styles at various levels

from traits.api import Enum, HasTraits, Str
from traitsui.api import Group, Item, View


class MixedStyles(HasTraits):
    first_name = Str()
    last_name = Str()

    department = Enum("Business", "Research", "Admin")
    position_type = Enum("Full-Time", "Part-Time", "Contract")

    traits_view = View(
         Group(
             Item(name='first_name'),
             Item(name='last_name'),
             Group(
                 Item(name='department'),
                 Item(name='position_type', style='custom'),
                 style='simple',
             ),
         ),
         title='Mixed Styles',
         style='readonly',
    )


ms = MixedStyles(first_name='Sam', last_name='Smith')
ms.configure_traits()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import unittest

from traits.api import HasTraits, Int, Str, Tuple
from traitsui.api import Item, View
from traits.testing.api import UnittestTools

from traitsui.tests._tools import (
    BaseTestMixin,
    create_ui,
    requires_toolkit,
    reraise_exceptions,
    ToolkitName,
)


class TupleEditor(HasTraits):
    """Dialog containing a Tuple of two Int's.
    """

    tup = Tuple(Int, Int, Str)

    traits_view = View(
        Item(label="Enter 4 and 6, then press OK"), Item("tup"), buttons=["OK"]
    )


class TestTupleEditor(BaseTestMixin, unittest.TestCase, UnittestTools):

    def setUp(self):
        BaseTestMixin.setUp(self)

    def tearDown(self):
        BaseTestMixin.tearDown(self)

    @requires_toolkit([ToolkitName.qt, ToolkitName.wx])
    def test_value_update(self):
        # Regression test for #179
        model = TupleEditor()
        with create_ui(model):
            with self.assertTraitChanges(model, "tup", count=1):
                model.tup = (3, 4, "nono")

    @requires_toolkit([ToolkitName.qt])
    def test_qt_tuple_editor(self):
        # Behavior: when editing the text of a tuple editor,
        # value get updated immediately.

        from pyface import qt

        val = TupleEditor()
        with reraise_exceptions(), create_ui(val) as ui:

            # the following is equivalent to clicking in the text control of
            # the range editor, enter a number, and clicking ok without
            # defocusing

            # text element inside the spin control
            lineedits = ui.control.findChildren(qt.QtGui.QLineEdit)
            lineedits[0].setFocus()
            lineedits[0].clear()
            lineedits[0].insert("4")
            lineedits[1].setFocus()
            lineedits[1].clear()
            lineedits[1].insert("6")
            lineedits[2].setFocus()
            lineedits[2].clear()
            lineedits[2].insert("fun")

            # if all went well, the tuple trait has been updated and its value
            # is (4, 6, "fun")
            self.assertEqual(val.tup, (4, 6, "fun"))


if __name__ == "__main__":
    # Executing the file opens the dialog for manual testing
    val = TupleEditor()
    val.configure_traits()
    print(val.tup)



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" General regression tests for various fixed bugs.
"""

import unittest

from traits.api import DelegatesTo, Event, HasTraits, Instance, Undefined
from traitsui.api import Editor, TextEditor

from traitsui.tests._tools import (
    BaseTestMixin,
)


class Parent(HasTraits):
    button = Event()


class Child(HasTraits):
    parent = Instance(Parent)
    button = DelegatesTo("parent")


class TestRegression(BaseTestMixin, unittest.TestCase):

    def setUp(self):
        BaseTestMixin.setUp(self)

    def tearDown(self):
        BaseTestMixin.tearDown(self)

    def test_editor_on_delegates_to_event(self):
        """ Make sure that DelegatesTo on Events passes Editor creation.
        """
        child = Child(parent=Parent())
        editor = Editor(
            None, factory=TextEditor(), object=child, name="button"
        )
        self.assertIs(editor.old_value, Undefined)

    def test_attribute_error(self):
        """ Make sure genuine AttributeErrors raise on Editor creation.
        """
        self.assertRaises(
            AttributeError,
            Editor,
            None,
            factory=TextEditor(),
            object=Parent(),
            name="not_a_trait",
        )
